// proc_details.cpp : implementation file
//

#include "stdafx.h"
#include "..\component.h"
#include "proc_details.h"

#include "..\linux_port\processor_network.h"

#include <map>
#include <list>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//externals
extern Graph grp;

using namespace std;

/** Static Helper Functions **/

//Check each group to make sure that there is not another task or series of tasks
//that have both an in and out data connection from the task.  If such a task is found
//then break it up into seperate groups so the problem is fixed 
static void AnalyzeAndFix(list<list<TaskType*> >& groups);

//Function returns true if the task has a data connection from a task outside of the 
//given group which then directly or indirectly has a data connection from a member
//of the same group.  checked parameter is so that the same task is not checked more than
//once
static bool ConnectedToSelf(TaskType* task, list<TaskType*>& original_group, map<TaskType*, bool> checked);

//recursive helper function for ConnectedToSelf function
static bool ConnectedToSelfRecurs(TaskType* task, list<TaskType*>& original_group, map<TaskType*, bool> checked);

//make a group from the tasks in the original_group starting from the given task and
//including every task that has a direct or indirect connection coming from it.  This
//group is put into the new_group variable, the checked variable is so that the same
//task is not checked more than once 
static void MakeGroupAfter(TaskType* task, list<TaskType*>& new_group, list<TaskType*>& original_group, map<TaskType*, bool> checked);

//remove all of the tasks that are in the new group from the old group and return the 
//position after the given task from the old group - if this task isn't found 
//old_group.end() is returned
static list<TaskType*>::iterator RemoveNewGroupFromOldGroup(TaskType* task, list<TaskType*>& old_group, list<TaskType*>& new_group);

void show_grouping(Processor* proc)
{
	proc_details window(proc);
	window.DoModal();			//display window
}

CString get_groups(Processor* proc_ptr)
{
	multimap<float, TaskType*>	period_grouping;
	list<list<TaskType*> >		groups;				//a list of the groups created
	multimap<float, TaskType*>::iterator task_it; 

	//group the tasks together by their periods
	for(list<TaskType*>::iterator group_it = proc_ptr->Tasks_Begin(); group_it != proc_ptr->Tasks_End(); ++group_it)
		period_grouping.insert(make_pair((*group_it)->Property()->Get_Period(), *group_it));

	task_it = period_grouping.begin();

	//group all of the tasks together by their connections and their periods
	while(task_it != period_grouping.end())
	{
		float period = task_it->first;
		list<list<TaskType*> > current_groups;	//list of groups for current period
		TaskType* task_ptr = task_it->second;

		//go through all of the tasks of the same period
		while(task_it != period_grouping.upper_bound(period))
		{
			task_ptr = task_it->second;
			FloodFill(task_ptr, current_groups, period_grouping);
			++task_it;
		}

		//add all of the groups that were just found to the overall list of groups
		for(list<list<TaskType*> >::iterator group_it = current_groups.begin(); group_it != current_groups.end(); ++group_it)
			groups.push_back(*group_it);

		task_ptr = task_it->second;
	}

	//merge the groups that are completely unconnected
	merge_unconnected(groups);

	//fix groups that need to be broken up
	AnalyzeAndFix(groups);

	return form_message(groups);
}

list<TaskType*> GetGroup(TaskType* task, list<list<TaskType*> > groups)
{
	list<TaskType*> empty;

	for(list<list<TaskType*> >::iterator group_it = groups.begin(); group_it != groups.end(); ++group_it)
		for(list<TaskType*>::iterator task_it = group_it->begin(); task_it != group_it->end(); ++task_it)
			if(task == *task_it) return *group_it;

	return empty;
}

void FloodFill(TaskType* task, list<list<TaskType*> >& groups, multimap<float, TaskType*> period_grouping)
{
	list<TaskType*> curr_group = GetGroup(task, groups);	//get the group the task is part of
	bool new_group = false;

	if(curr_group.empty())	//task was not part of a previous group - start a new group
	{
		new_group = true;
		curr_group.push_back(task);
	}

	//add all of the directly connected tasks with the same period to the group
	for(list<ConnectionType*>::iterator conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_In_Data_Edges_End(); ++conn_it)
	{
		ConnectionType* conn_ptr = *conn_it;
		TaskType* from_task = conn_ptr->Get_From();

		if(from_task->Property()->Get_Period() == task->Property()->Get_Period() && 
			OnProc(from_task, period_grouping))
		{
			list<TaskType*> froms_group = GetGroup(from_task, groups);
			if(!froms_group.empty())								//is this task already part of another group?
				merge(curr_group, froms_group, groups);
			else if(from_task && !onList(from_task, curr_group))	//is the task alreay in this group?
				curr_group.push_back(from_task);
		}
	}

	//add a new group or update the old group
	if(new_group)
		groups.push_back(curr_group);
	else
		update(curr_group, groups);
}

bool onList(TaskType* task, list<TaskType*> task_list)
{
	for(list<TaskType*>::iterator it = task_list.begin(); it != task_list.end(); ++it)
		if(task == *it) return true;

	return false;
}

list<list<TaskType*> >::iterator merge(list<TaskType*>& absorb_group, list<TaskType*>& delete_group, list<list<TaskType*> >& groups)
{
	list<list<TaskType*> >::iterator next_it;

	//if the groups are equal return the end iterator
	if(absorb_group.front() == delete_group.front())
		return groups.end();

	//copy the group that is going to be deleted
	list<TaskType*> copy_delete = delete_group;

	//find the delete group and erase it
	for(list<list<TaskType*> >::iterator group_it = groups.begin(); group_it != groups.end(); ++group_it)
		if(delete_group.front() == group_it->front()) 
		{
			next_it = groups.erase(group_it);
			break;
		}

	//add in the tasks of the deleted group
	for(list<TaskType*>::iterator it = copy_delete.begin(); it != copy_delete.end(); ++it)
		absorb_group.push_back(*it);

	return next_it;
}

bool OnProc(TaskType* task, multimap<float, TaskType*> period_grouping)
{
	float period = task->Property()->Get_Period();

	for(multimap<float, TaskType*>::iterator it = period_grouping.lower_bound(period); it != period_grouping.upper_bound(period); ++ it)
		if(task == it->second) return true;

	return false;
}

void update(list<TaskType*> update_group, list<list<TaskType*> >& groups)
{
	//find the group
	for(list<list<TaskType*> >::iterator group_it = groups.begin(); group_it != groups.end(); ++group_it)
	{
		if(update_group.front() == group_it->front())
		{
			*group_it = update_group;	//found the group replace it with the given group
			return;
		}
	}
}

CString form_message(list<list<TaskType*> > groups)
{
	CString msg;
	CString current;

	if(groups.empty())
		return CString("Processory empty");

	int group_num = 1;
	char group_num_cstr[10];	//cstring for the group number
	char period[20];			//cstring for the period

	for(list<list<TaskType*> >::iterator group_it = groups.begin(); group_it != groups.end(); ++group_it)
	{
		sprintf(group_num_cstr, "%i", group_num++);
		sprintf(period, "%f", group_it->front()->Property()->Get_Period());

		msg += CString("Group ") + group_num_cstr;
		msg += CString("\tPeriod: ") + period + " ms";
		msg += "\r\n------------------------------------------------------------------------------------------\r\n";

		for(list<TaskType*>::iterator task_it = group_it->begin(); task_it != group_it->end(); ++task_it)
		{
			if(*task_it)
			{
				//trial and error used to find the values below to try to even the tabs
				//feel free to change to make the output look better
				string s;
				s =  (*task_it)->Property()->Get_Name() + '\t';
				if(s.length() < 9)
					s += string("\t");
				if(s.length() < 20)
					s += string("\t");

				double utilization = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();
				current.Format("%s  Utilization:   %f\r\n", s.c_str(), utilization);
				msg += current;
			}
		}

		msg += "\r\n";
	}

	return msg;
}

void merge_unconnected(list<list<TaskType*> >& groups)
{
	list<list<TaskType*> >::iterator group1_it = groups.begin();;
	list<list<TaskType*> >::iterator group2_it;

	//go through each group
	while(group1_it != groups.end())
	{
		//go through each group after group1
		for(group2_it = group1_it; group2_it != groups.end(); ++group2_it)
		{
			if(group2_it == group1_it)	//move foreward once after initilization
				continue;

			if(not_connected(*group1_it, *group2_it))
			{
				group2_it->push_back(NULL);		//NULL used to diferentiate between different connected groups
				group1_it = merge(*group2_it, *group1_it, groups);
				break;
			}
		}

		if(group2_it == groups.end())	//could not merge group1 with any other groups
			++group1_it;				//need to increment the iterator
	}
}

bool not_connected(list<TaskType*> group1, list<TaskType*> group2)
{
	map<TaskType*, bool> visited;
	list<TaskType*>::iterator g1_it = group1.begin();
	list<TaskType*>::iterator g2_it;

	//go through each task in group1
	while(g1_it != group1.end())
	{
		//go through each task in group2
		g2_it = group2.begin();
		while(g2_it != group2.end())
		{
			//clear out and initialize visited mapping
			visited.clear();
			for(list<TaskType*>::iterator task_it = group2.begin(); task_it != group2.end(); ++task_it)
				visited[*task_it] = false;

			//are these two tasks connected
			if(indirectly_connected(*g1_it, *g2_it, visited))
				return false;

			//move iterator foreward until the end or NULL is found (NULL means 
			//that a different group of connected nodes was added to the group) 
			while(g2_it != group2.end() && *g2_it != NULL)
				g2_it++;
			if(*g2_it == NULL)	//move iterator past NULL
				g2_it++;
		}

		//move iterator foreward until the end or NULL is found (NULL means 
		//that a different group of connected nodes was added to the group)
		while(g1_it != group1.end() && *g1_it != NULL)
			g1_it++;
		if(*g1_it == NULL)	//move iterator past NULL
			g1_it++;
	}

	return true;
}

bool indirectly_connected(TaskType* task, TaskType* current, map<TaskType*, bool>& visited)
{
	if(all_visited(visited))	//all of the tasks have been visited - can't check any others
		return false;

	visited[current] = true;	//mark this task as visited

	//go through each in connection
	for(list<ConnectionType*>::iterator conn_it = current->Get_In_Data_Edges_Begin(); conn_it != current->Get_In_Data_Edges_End(); ++conn_it)
	{
		TaskType* from = (*conn_it)->Get_From();
		if((*conn_it)->Get_From() == task)			//task found
			return true;

		if(!visited[(*conn_it)->Get_From()])		//if this task hasn't been visited
		{											//visit it (recursive)
			if(indirectly_connected(task, (*conn_it)->Get_From(), visited))
				return true;
		}
	}

	//go through each out connection
	for(conn_it = current->Get_Out_Data_Edges_Begin(); conn_it != current->Get_Out_Data_Edges_End(); ++conn_it)
	{
		if((*conn_it)->Get_To() == task)			//task found
			return true;

		if(!visited[(*conn_it)->Get_To()])			//if this task hasn't been visited
		{											//visit it (recursive)
			if(indirectly_connected(task, (*conn_it)->Get_To(), visited))
				return true;
		}
	}

	return false;
}

bool all_visited(map<TaskType*, bool>& visited)
{
	for(map<TaskType*, bool>::iterator it = visited.begin(); it != visited.end(); ++it)
	{
		TaskType* task = it->first;
		if(!it->second)
			return false;
	}

	return true;
}

void AnalyzeAndFix(list<list<TaskType*> >& groups)
{
	list<list<TaskType*> >::iterator group_it;
	list<TaskType*>::iterator task_it;
	list<TaskType*>::iterator checked_task_it;
	map<TaskType*, bool> checked;

	//go through each group
	for(group_it = groups.begin(); group_it != groups.end(); ++group_it)
	{
		//initialize the checked tasks mapping
		checked.clear();
		for(checked_task_it = grp.Node_Begin(); checked_task_it != grp.Node_End(); ++checked_task_it)
			checked[*checked_task_it] = false;

		task_it = group_it->begin();
		while(task_it != group_it->end())
		{
			TaskType* task = *task_it;
	
			//only wory about task if it is non null and has more then 1 data connection
			if(!task || task->Get_Number_Of_In_Data_Connection() < 2)
			{
				++task_it;
				continue;
			}

			if(ConnectedToSelf(task, *group_it, checked))
			{
				//reset checked variable for just the variables in the group
				checked.clear();
				for(checked_task_it = group_it->begin(); checked_task_it != group_it->end(); ++checked_task_it)
					checked[*checked_task_it] = false;

				list<TaskType*> new_group;
				MakeGroupAfter(task, new_group, *group_it, checked);				//create the new group
				task_it = RemoveNewGroupFromOldGroup(task, *group_it, new_group);	//remove the members of the 
																					//new group from the old group
				groups.push_back(new_group);
			}
			else
				++task_it;
		}
	}

	return;
}

bool ConnectedToSelf(TaskType* task, list<TaskType*>& original_group, map<TaskType*, bool> checked)
{
	list<ConnectionType*>::iterator conn_it;
	list<TaskType*>::iterator task_it;

	//go through each in data connection for the task
	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_In_Data_Edges_End(); ++conn_it)
	{
		TaskType* from = (*conn_it)->Get_From();

		//check to see if the task from is in the original group
		for(task_it = original_group.begin(); task_it != original_group.end(); ++task_it)
			if(*task_it == from) break;

		//task from is not in the original group - check to see if the original can be reached
		if(task_it == original_group.end())
		{
			if(ConnectedToSelfRecurs(from, original_group, checked))
				return true;
		}
	}

	return false;
}

bool ConnectedToSelfRecurs(TaskType* task, list<TaskType*>& original_group, map<TaskType*, bool> checked)
{
	list<ConnectionType*>::iterator conn_it;
	list<TaskType*>::iterator task_it;

	if(!task->Get_Number_Of_In_Data_Connection())
		return false;

	//was this task already checked
	if(checked[task])
		return false;

	//mark task as checked
	checked[task] = true;

	//go through each in data connection for the task
	for(conn_it = task->Get_In_Data_Edges_Begin(); conn_it != task->Get_In_Data_Edges_End(); ++conn_it)
	{
		TaskType* from = (*conn_it)->Get_From();

		//return true if task from is in the original group
		for(task_it = original_group.begin(); task_it != original_group.end(); ++task_it)
			if(*task_it == from) return true;

		//task from wasn't in the original group - keep on looking
		if(ConnectedToSelfRecurs(from, original_group, checked))
			return true;
	}

	return false;
}

void MakeGroupAfter(TaskType* task, list<TaskType*>& new_group, list<TaskType*>& original_group, map<TaskType*, bool> checked)
{
	list<ConnectionType*>::iterator conn_it;
	list<TaskType*>::iterator task_it;

	//was this task already checked
	if(checked[task])
		return;

	//mark as checked
	checked[task] = true;

	new_group.push_back(task);

	if(!task->Get_Number_Of_Out_Data_Connection())
		return;

	//go through each out data connections
	for(conn_it = task->Get_Out_Data_Edges_Begin(); conn_it != task->Get_Out_Data_Edges_End(); ++conn_it)
	{
		TaskType* to = (*conn_it)->Get_To();

		//check to see if this task is part of the original group
		for(task_it = original_group.begin(); task_it != original_group.end(); ++task_it)
		{
			if(*task_it == to)	//task found
			{
				if(!onList(to, new_group))	//only worry about the task if it isn't already part of the new group
					MakeGroupAfter(to, new_group, original_group, checked);
				break;
			}
		}
	}
}

list<TaskType*>::iterator RemoveNewGroupFromOldGroup(TaskType* task, list<TaskType*>& old_group, list<TaskType*>& new_group)
{
	list<TaskType*>::iterator new_it;
	list<TaskType*>::iterator old_it;
	list<TaskType*>::iterator ret_it = old_group.end();

	//go through each member of the new group
	for(new_it = new_group.begin(); new_it != new_group.end(); ++new_it)
	{
		//go through each member of the old group
		for(old_it = old_group.begin(); old_it != old_group.end(); ++old_it)
			if(*old_it == *new_it)	//found the task
			{
				if(*old_it = task)	//is this the task that was given - if so store next iterator position
					ret_it = old_group.erase(old_it);
				else
					old_group.erase(old_it);
				break;
			}
	}

	return ret_it;
}

/////////////////////////////////////////////////////////////////////////////
// proc_details dialog


proc_details::proc_details(Processor* proc, CWnd* pParent /*=NULL*/)
	: CDialog(proc_details::IDD, pParent)
{
	proc_ptr = proc;
	initialized = false;

	//{{AFX_DATA_INIT(proc_details)
	//}}AFX_DATA_INIT
}

void proc_details::initialize()
{
	if(!proc_ptr)
		return;

	char id[20];
	CString type;
	char bound[20];
	char utilization[20];

	//find the type of processor that the given processor is
	switch(proc_ptr->Property()->Get_Type())
	{
	case 0:
		type = "MPC";
		break;
	case 1:
		type = "INTEL";
		break;
	default:
		type = "UNKNOWN";
		break;
	}

	sprintf(id, "%i", proc_ptr->Property()->Get_CAN_ID());
	sprintf(bound, "%f", proc_ptr->Property()->Get_Bound());
	sprintf(utilization, "%f", proc_ptr->Property()->Get_Bound() - proc_ptr->Property()->Get_Available_Space());

	//fill in the text fields
	m_name.SetWindowText(proc_ptr->Property()->Get_Name());
	m_id.SetWindowText(id);
	m_type.SetWindowText(type);
	m_bound.SetWindowText(bound);
	m_utilization.SetWindowText(utilization);
	m_groups.SetWindowText(get_groups(proc_ptr));

	initialized = true;
}

void proc_details::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(proc_details)
	DDX_Control(pDX, IDC_EDIT1, m_name);
	DDX_Control(pDX, IDC_EDIT2, m_groups);
	DDX_Control(pDX, IDC_EDIT3, m_id);
	DDX_Control(pDX, IDC_EDIT4, m_type);
	DDX_Control(pDX, IDC_EDIT5, m_bound);
	DDX_Control(pDX, IDC_EDIT6, m_utilization);
	//}}AFX_DATA_MAP

	if(!initialized)
		initialize();
}


BEGIN_MESSAGE_MAP(proc_details, CDialog)
	//{{AFX_MSG_MAP(proc_details)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// proc_details message handlers
