#if !defined(AFX_PROC_DETAILS_H__3D051464_53AD_48AE_BFE6_F77D9E361413__INCLUDED_)
#define AFX_PROC_DETAILS_H__3D051464_53AD_48AE_BFE6_F77D9E361413__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
// proc_details.h : header file
//

#include "Resource.h"
#include <list>
#include <map>

class Processor;
class TaskType;

//pop a window showing the details of the given processor
void show_grouping(Processor* proc);

//get a string that groups the tasks on the processor together
//Currently the groups are formed between tasks that are connected, on the same processor,
//(the processor that was given), and have the same period. 
CString get_groups(Processor* proc_ptr);

//Returns a list that the given task is part of from the given groups, returns an empty
//list if the task is not part of any of the groups 
std::list<TaskType*> GetGroup(TaskType* task, std::list<std::list<TaskType*> > groups);

//Puts the given task in a group if it is not already part of another.  Then puts all of
//the tasks that are connected to it and have the same period into this group.  If any
//of its connected tasks are already part of another group then these two groups are
//merged together
void FloodFill(TaskType* task, std::list<std::list<TaskType*> >& groups, std::multimap<float, TaskType*> period_mapping);

//Moves all of the tasks from the delete group to the absorb group and removes the delete 
//group from the given list of groups.  Returns the iterator position for the list of
//groups that is just after the position of the delete group 
std::list<std::list<TaskType*> >::iterator merge(std::list<TaskType*>& absorb_group, std::list<TaskType*>& delete_group, std::list<std::list<TaskType*> >& groups);

//Returns true if the given task is on the given processor
bool OnProc(TaskType* task, std::multimap<float, TaskType*> period_mapping);

//Switch the group that has the same begining task in the list of groups with the given
//group
void update(std::list<TaskType*> update_group, std::list<std::list<TaskType*> >& groups);

//Returns a string that is formatted to display the given list of groups
CString form_message(std::list<std::list<TaskType*> > groups);

//Return true if the given task is on the given list
bool onList(TaskType* task, std::list<TaskType*> task_list);

/*** The following four functions are currently not used (to use them uncomment the
 *** call to merge_unconnected() in get_groups() ).  However, I am leaving them in 
 *** since they may be usefull for a more complex task grouping on a processor.
 ***/

//Merge all of the completely unconnected groups together 
void merge_unconnected(std::list<std::list<TaskType*> >& groups);

//Returns true if there are no connections at all between any tasks in group1 to any
//tasks in group2
bool not_connected(std::list<TaskType*> group1, std::list<TaskType*> group2);

//Returns true if all of the entries in visited have a value of true - used to determine
//if all the tasks in the graph have been examined
bool all_visited(std::map<TaskType*, bool>& visited);

//Returns true if the given TaskType task is connected in some way with the given TaskType
//current, the map visited is used for recursive purposes to keep track of the tasks that
//have been examined
bool indirectly_connected(TaskType* task, TaskType* current, std::map<TaskType*, bool>& visited);


/////////////////////////////////////////////////////////////////////////////
// proc_details dialog

class proc_details : public CDialog
{
// Construction
public:
	proc_details(Processor* proc = NULL, CWnd* pParent = NULL);   // standard constructor

	void initialize();

// Dialog Data
	//{{AFX_DATA(proc_details)
	enum { IDD = IDD_DIALOG3 };
	CEdit	m_name;
	CEdit	m_groups;
	CEdit	m_id;
	CEdit	m_type;
	CEdit	m_bound;
	CEdit	m_utilization;
	//}}AFX_DATA


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(proc_details)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:

	Processor* proc_ptr;
	bool initialized;
	// Generated message map functions
	//{{AFX_MSG(proc_details)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_PROC_DETAILS_H__3D051464_53AD_48AE_BFE6_F77D9E361413__INCLUDED_)
