// rand_sys_gen_dialog.cpp : implementation file
//

#include "stdafx.h"
#include "..\component.h"
#include "../linux_port/tools.h"
#include "..\linux_port\graph.h"
#include "..\linux_port\system.h"
#include "../Random System Generator/gen_sys.h"
#include "rand_sys_gen_dialog.h"

#include <algorithm>
#include <stdio.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// RandSysGenerator dialog

//static helper functions

//set the both the param and the text in the window of the edit box to be the value given
static void set_vars(int value, int& param, CEdit* edit); 
static void set_vars(float value, float& param, CEdit* edit);

//put the value that is in the window of the edit box into the param
static void parse(int& param, CEdit* edit);
static void parse(float& param, CEdit* edit);

//determine what the error is and display the appropriate message box - returns true if
//the code where it was called should continue, false otherwise
static bool handle_error(Error<CString> e);

//globals
bool use_gme_hw = false;	//set to true when the hardware in GME should be used instead 
							//of randomly generated

//externals
extern Graph rand_grp;		//graph and system generated randomly
extern System rand_sys;
extern Graph GME_grp;		//graph and system from GME model
extern System GME_sys;

RandSysGenerator::RandSysGenerator(CWnd* pParent /*=NULL*/)
	: CDialog(RandSysGenerator::IDD, pParent)
{
	//{{AFX_DATA_INIT(RandSysGenerator)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

	max_procs = -1;
}


void RandSysGenerator::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(RandSysGenerator)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(RandSysGenerator, CDialog)
	//{{AFX_MSG_MAP(RandSysGenerator)
	ON_BN_CLICKED(IDC_USE_GME_HW_CB, OnUseGMEhw)
	ON_BN_CLICKED(IDC_GEN_SYS_BUTTON, OnGenerateRandSys)
	ON_BN_CLICKED(IDC_VIEW_SYS_BUTTON, OnViewSys)
	ON_BN_CLICKED(IDC_SAVE_SYS_BUTTON, OnSaveSys)
	ON_BN_CLICKED(IDC_CREATE_NEW_SYS_RB, OnGenNewSysRb)
	ON_BN_CLICKED(IDC_ADD_TO_SYS_RB, OnAppendSysRb)

	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// RandSysGenerator message handlers

BOOL RandSysGenerator::OnInitDialog()
{
	CDialog::OnInitDialog();

	use_gme_hw_cb			= static_cast<CButton*>(GetDlgItem(IDC_USE_GME_HW_CB));
	view_sys_but			= static_cast<CButton*>(GetDlgItem(IDC_VIEW_SYS_BUTTON));
	save_sys_but			= static_cast<CButton*>(GetDlgItem(IDC_SAVE_SYS_BUTTON));
	create_new_sys_rb		= static_cast<CButton*>(GetDlgItem(IDC_CREATE_NEW_SYS_RB));
	add_to_sys_rb			= static_cast<CButton*>(GetDlgItem(IDC_ADD_TO_SYS_RB));

	max_procs_edit			= static_cast<CEdit*>(GetDlgItem(IDC_NUM_PROCS_MAX_EDIT));
	min_procs_edit			= static_cast<CEdit*>(GetDlgItem(IDC_NUM_PROCS_MIN_EDIT));
	max_bound_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MAX_BOUND_EDIT));		
	min_bound_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MIN_BOUND_EDIT));
	max_comps_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MAX_NUM_COMPS_EDIT));
	min_comps_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MIN_NUM_COMPS_EDIT));
	max_comp_util_edit		= static_cast<CEdit*>(GetDlgItem(IDC_MAX_COMP_UTIL_EDIT));
	min_comp_util_edit		= static_cast<CEdit*>(GetDlgItem(IDC_MIN_COMP_UTIL_EDIT));
	max_period_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MAX_PERIOD_EDIT));
	min_period_edit			= static_cast<CEdit*>(GetDlgItem(IDC_MIN_PERIOD_EDIT));
	max_input_comps_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MAX_INPUT_COMPS));
	min_input_comps_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MIN_INPUT_COMPS));
	max_output_comps_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MAX_OUTPUT_COMPS_EDIT));
	min_output_comps_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MIN_OUTPUT_COMPS_EDIT));
	max_output_degree_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MAX_COMP_OUTPUT_DEGREE_EDIT));
	min_output_degree_edit	= static_cast<CEdit*>(GetDlgItem(IDC_MIN_COMP_OUTPUT_DEGREE_EDIT));
	max_msg_size_edit		= static_cast<CEdit*>(GetDlgItem(IDC_MAX_MSG_SIZE_EDIT));
	min_msg_size_edit		= static_cast<CEdit*>(GetDlgItem(IDC_MIN_MSG_SIZE_EDIT));

	if(max_procs == -1)		//this is the first time the random system generator was run - use default values
	{
		set_vars(4,   max_procs,			max_procs_edit);
		set_vars(1,   min_procs,			min_procs_edit);
		set_vars(1,   max_bound,			max_bound_edit);
		set_vars(.5,  min_bound,			min_bound_edit);
		set_vars(50,  max_comps,			max_comps_edit);
		set_vars(10,  min_comps,			min_comps_edit);
		set_vars(.3,  max_comp_util,     max_comp_util_edit);
		set_vars(.01, min_comp_util,     min_comp_util_edit);
		set_vars(20,  max_period,		max_period_edit);
		set_vars(10,  min_period,		min_period_edit);
		set_vars(5,   max_input_comps,   max_input_comps_edit);
		set_vars(1,   min_input_comps,   min_input_comps_edit);
		set_vars(5,   max_output_comps,  max_output_comps_edit);
		set_vars(1,   min_output_comps,  min_output_comps_edit);
		set_vars(5,   max_output_degree, max_output_degree_edit);
		set_vars(1,   min_output_degree, min_output_degree_edit);
		set_vars(200, max_msg_size,      max_msg_size_edit);
		set_vars(1,   min_msg_size,      min_msg_size_edit);
	}
	else				//this is not the first time the random system generator was run - use previous values
	{
		set_vars(max_procs,		   max_procs,			max_procs_edit);
		set_vars(min_procs,		   min_procs,			min_procs_edit);
		set_vars(max_bound,         max_bound,			max_bound_edit);
		set_vars(min_bound,         min_bound,			min_bound_edit);
		set_vars(max_comps,         max_comps,			max_comps_edit);
		set_vars(min_comps,         min_comps,			min_comps_edit);
		set_vars(max_comp_util,     max_comp_util,		max_comp_util_edit);
		set_vars(min_comp_util,     min_comp_util,		min_comp_util_edit);
		set_vars(max_period,        max_period,			max_period_edit);
		set_vars(min_period,        min_period,			min_period_edit);
		set_vars(max_input_comps,   max_input_comps,		max_input_comps_edit);
		set_vars(min_input_comps,   min_input_comps,		min_input_comps_edit);
		set_vars(max_output_comps,  max_output_comps,	max_output_comps_edit);
		set_vars(min_output_comps,  min_output_comps,	min_output_comps_edit);
		set_vars(max_output_degree, max_output_degree,	max_output_degree_edit);
		set_vars(min_output_degree, min_output_degree,	min_output_degree_edit);
		set_vars(max_msg_size,      max_msg_size,		max_msg_size_edit);
		set_vars(min_msg_size,      min_msg_size,		min_msg_size_edit);
	}

	use_gme_hw = false;	//initially set that the hardware given in the GME model is
						//not used

	create_new_sys_rb->SetCheck(BST_CHECKED);	//check the radio button that says to
												//create a new system

	//is there already a random system generated
	if(rand_grp.Get_Number_Of_Nodes())
	{
		view_sys_but->EnableWindow();		//enable the view and save buttons
		save_sys_but->EnableWindow();

		create_new_sys_rb->EnableWindow();	//enable the radio buttons for creating a new system
		add_to_sys_rb->EnableWindow();		//or adding to the current system
	}

	return TRUE;
}

void RandSysGenerator::OnUseGMEhw()
{
	use_gme_hw = !use_gme_hw;	//toggle the value

	if(!use_gme_hw)				//if we are not using the GME hw enable the hw fields
	{
		max_procs_edit->EnableWindow();
		min_procs_edit->EnableWindow();
		max_bound_edit->EnableWindow();
		min_bound_edit->EnableWindow();
	}	
	else						//if we are using the GME hw disable the hw fields
	{
		max_procs_edit->EnableWindow(FALSE);
		min_procs_edit->EnableWindow(FALSE);
		max_bound_edit->EnableWindow(FALSE);
		min_bound_edit->EnableWindow(FALSE);
	}
}

void RandSysGenerator::OnGenerateRandSys() 
{
	try{
		parse_fields();				//read in the fields
	}catch(Error<CString> e)
	{
		if(!handle_error(e))		//handle any errors and return if necessary
			return;
	}

	RandSysParams param;
	param.max_procs = max_procs;
	param.min_procs = min_procs;
	param.max_bound = max_bound;
	param.min_bound = min_bound;
	param.max_comps = max_comps;
	param.min_comps = min_comps;
	param.max_util = max_comp_util;
	param.min_util = min_comp_util;
	param.max_period = max_period;
	param.min_period = min_period;
	param.max_input_comps = max_input_comps;
	param.min_input_comps = min_input_comps;
	param.max_output_comps = max_output_comps;
	param.min_output_comps = min_output_comps;
	param.max_output_degree = max_output_degree;
	param.min_output_degree = min_output_degree;
	param.max_msg_size = max_msg_size;
	param.min_msg_size = min_msg_size;
	param.use_GME_hw = use_gme_hw;
	param.append_sys = add_to_sys_rb->GetCheck() == BST_CHECKED;

	//generate the system
	GenerateRandSys(param);
	
	view_sys_but->EnableWindow();	//enable the view and save buttons
	save_sys_but->EnableWindow();

	create_new_sys_rb->EnableWindow();	//enable the radio buttons for creating a new system
	add_to_sys_rb->EnableWindow();		//or adding to the current system
	
}

void RandSysGenerator::parse_fields()
{
	Error<CString> e_big;
	Error<CString> e_small;
	Error<CString> e_switch;

	e_big.type    = TOO_BIG_NO_DISPLAY;
	e_small.type  = TOO_SMALL_NO_DISPLAY;
	e_switch.type = SWAP_VALUES;

	//parse all of the individual fields
	parse(max_procs, max_procs_edit);
	parse(min_procs, min_procs_edit);
	parse(max_bound, max_bound_edit);
	parse(min_bound, min_bound_edit);
	parse(max_comps, max_comps_edit);
	parse(min_comps, min_comps_edit);
	parse(max_comp_util, max_comp_util_edit);
	parse(min_comp_util, min_comp_util_edit);
	parse(max_period, max_period_edit);
	parse(min_period, min_period_edit);
	parse(max_input_comps, max_input_comps_edit);
	parse(min_input_comps, min_input_comps_edit);
	parse(max_output_comps, max_output_comps_edit);
	parse(min_output_comps, min_output_comps_edit);
	parse(max_output_degree, max_output_degree_edit);
	parse(min_output_degree, min_output_degree_edit);
	parse(max_msg_size, max_msg_size_edit);
	parse(min_msg_size, min_msg_size_edit);

	if(float(max_bound) > 1.000001)	//if the bound is above 1 - modified for errors with doubles
	{
		if(IDNO == AfxMessageBox("Warning: The maximum bound is greater than 1.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_big;
	}
	else if(float(max_bound) < 0)
	{
		if(IDNO == AfxMessageBox("Warning: The maximum bound is less than 0.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_small;
	}

	if(float(min_bound) > 1.000001) //if the bound is above 1 - modified for errors with doubles
	{
		if(IDNO == AfxMessageBox("Warning: The minimum bound is greater than 1.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_big;
	}
	else if(float(min_bound) < 0)
	{
		if(IDNO == AfxMessageBox("Warning: The minimum bound is less than 0.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_small;
	}

	if(float(max_comp_util) > 1.000001) //if the utilization is above 1 - modified for errors with doubles
	{
		if(IDNO == AfxMessageBox("Warning: The maximum component utilization is greater than 1.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_big;
	}
	else if(float(max_comp_util) < 0)
	{
		if(IDNO == AfxMessageBox("Warning: The maximum component utilization is less than 0.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_small;
	}

	if(float(min_comp_util) > 1.000001) //if the utilization is above 1 - modified for errors with doubles
	{
		if(IDNO == AfxMessageBox("Warning: The minimum component utilization is greater than 1.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_big;
	}
	else if(float(min_comp_util) < 0)
	{
		if(IDNO == AfxMessageBox("Warning: The minimum component utilization is less than 0.  Do you wish to continue?", MB_YESNO | MB_ICONEXCLAMATION))
			throw e_small;
	}

	CString too_small;

	//Make sure that the fields that need to be at least 0 are not less than 0
	if(max_procs < 0)
		too_small = "Maximun number of processors";
	if(min_procs < 0)
		too_small = "Minimun number of processors";
	if(max_comps < 0)
		too_small = "Maximun number of components";
	if(min_comps < 0)
		too_small = "Minimun number of components";
	if(max_input_comps < 0)
		too_small = "Maximun number of input components";
	if(min_input_comps < 0)
		too_small = "Minimun number of input components";
	if(max_output_comps < 0)
		too_small = "Maximun number of output components";
	if(min_output_comps < 0)
		too_small = "Minimun number of output components";
	if(max_output_degree < 0)
		too_small = "Maximun component output degree";
	if(min_output_degree < 0)
		too_small = "Minimun component output degree";
	if(max_msg_size < 0)
		too_small = "Maximun message size";
	if(min_msg_size < 0)
		too_small = "Minimun message size";
	if(float(min_period) < 0)
		too_small = "Minimun period";
	if(float(max_period) < 0)
		too_small = "Maximun period";

	if(too_small.GetLength() != 0)
	{
		e_small.type = TOO_SMALL_DISPLAY;
		e_small.msg = "Error: " + too_small + " is less than 0.";
		throw e_small;
	}

	too_small = "";

	//Make sure that none of the fields that need to be at least 1 are not 0
	if(max_input_comps == 0)
		too_small = "Maximun number of input components";
	if(min_input_comps == 0)
		too_small = "Minimun number of input components";
	if(max_output_comps == 0)
		too_small = "Maximun number of output components";
	if(min_output_comps == 0)
		too_small = "Minimun number of output components";
	if(max_output_degree == 0)
		too_small = "Maximun component output degree";
	if(min_output_degree == 0)
		too_small = "Minimun component output degree";

	if(too_small.GetLength() != 0)
	{
		e_small.type = TOO_SMALL_DISPLAY;
		e_small.msg = "Error: " + too_small + " must be greater than 0.";
		throw e_small;
	}

	//make sure that all max's are >= all min's - if not swap the values
	if(max_procs < min_procs)
	{
		e_switch.templated_list1.push_back("Number of processors");
		swap(max_procs, min_procs);
	}
	if(max_bound < min_bound)
	{
		e_switch.templated_list1.push_back("Processor bounds");
		swap(max_bound, min_bound);
	}
	if(max_comps < min_comps)
	{
		e_switch.templated_list1.push_back("Number of components");
		swap(max_comps, min_comps);
	}
	if(max_comp_util < min_comp_util)
	{
		e_switch.templated_list1.push_back("Component utilization");
		swap(max_comp_util, min_comp_util);
	}
	if(max_period < min_period)
	{
		e_switch.templated_list1.push_back("Component periods");
		swap(max_period, min_period);
	}
	if(max_input_comps < min_input_comps)
	{
		e_switch.templated_list1.push_back("Number of input components");
		swap(max_input_comps, min_input_comps);
	}
	if(max_output_comps < min_output_comps)
	{
		e_switch.templated_list1.push_back("Number of output components");
		swap(max_output_comps, min_output_comps);
	}
	if(max_output_degree < min_output_degree)
	{
		e_switch.templated_list1.push_back("Component output degree");
		swap(max_output_degree, min_output_degree);
	}
	if(max_msg_size < min_msg_size)
	{
		e_switch.templated_list1.push_back("Message size");
		swap(max_msg_size, min_msg_size);
	}

	if(e_switch.templated_list1.size())
		throw e_switch;
}

void RandSysGenerator::OnViewSys()
{
	if(use_gme_hw)
	{
		if(GME_sys.GetDescription() == "")
			GenerateDescription(&GME_sys);
	}
	else
	{
		if(rand_sys.GetDescription() == "")
			GenerateDescription(&rand_sys);
	}

	if(rand_grp.GetDescription() == "")
		GenerateDescription(&rand_grp);

	if(use_gme_hw)
		view_system.SetText(GME_sys.GetDescription() + rand_grp.GetDescription());
	else
		view_system.SetText(rand_sys.GetDescription() + rand_grp.GetDescription());

	view_system.DoModal();
}

void RandSysGenerator::OnSaveSys()
{
	if(use_gme_hw)
		StoreSystem(GME_sys, rand_grp);
	else
		StoreSystem(rand_sys, rand_grp);
}

void RandSysGenerator::OnGenNewSysRb()
{
	use_gme_hw_cb->EnableWindow();		//enable the use GME hardware check box

	if(!use_gme_hw)				//if we are not using the GME hw enable the hw fields
	{
		use_gme_hw_cb->SetCheck(BST_UNCHECKED);
		max_procs_edit->EnableWindow();
		min_procs_edit->EnableWindow();
		max_bound_edit->EnableWindow();
		min_bound_edit->EnableWindow();
	}	
	else						//if we are using the GME hw disable the hw fields
	{
		use_gme_hw_cb->SetCheck(BST_CHECKED);
		max_procs_edit->EnableWindow(FALSE);
		min_procs_edit->EnableWindow(FALSE);
		max_bound_edit->EnableWindow(FALSE);
		min_bound_edit->EnableWindow(FALSE);
	}
}

void RandSysGenerator::OnAppendSysRb()
{
	if(!rand_sys.Get_Number_Of_Nets() && rand_grp.Get_Number_Of_Nodes())
	{
		use_gme_hw_cb->SetCheck(BST_CHECKED);
		max_procs_edit->EnableWindow(FALSE);
		min_procs_edit->EnableWindow(FALSE);
		max_bound_edit->EnableWindow(FALSE);
		min_bound_edit->EnableWindow(FALSE);
	}
	else
	{
		use_gme_hw_cb->SetCheck(BST_UNCHECKED);
		max_procs_edit->EnableWindow();
		min_procs_edit->EnableWindow();
		max_bound_edit->EnableWindow();
		min_bound_edit->EnableWindow();
	}

	use_gme_hw_cb->EnableWindow(FALSE);	//disable the use GME hardware check box
}

void set_vars(int value, int& param, CEdit* edit)
{
	char buff[20];

	param=value;
	sprintf(buff, "%d", param);

	edit->SetWindowText(buff);
}

void set_vars(float value, float& param, CEdit* edit)
{
	char buff[20];

	param=value;
	sprintf(buff, "%.2f", param);

	edit->SetWindowText(buff);
}

void parse(int& param, CEdit* edit)
{
	char buff[20];
	edit->GetWindowText(buff, 20);
	sscanf(buff, "%d", &param);
}

static void parse(float& param, CEdit* edit)
{
	char buff[20];
	edit->GetWindowText(buff, 20);

	bool decimal = false;
	bool error = false;

	//parse buff to make sure it is a number
	for(int i = 0; i < 20; i++)
	{
		if(buff[i] == 0)
			break;

		if(buff[i] == '-')
		{
			if(i != 0)
			{
				error = true;
				break;
			}
		} 
		else if(buff[i] == '.')
		{
			if(!decimal)
				decimal = true;
			else
			{
				error = true;
				break;
			}
		}
		else if(!isdigit(buff[i]))
		{
			error = true;
			break;
		}
	}

	if(error)
	{
		Error<CString> e;
		e.msg = CString(buff) + " does not evaluate to a number.";
		e.type = NOT_A_NUMBER;
		throw e;
	}

	sscanf(buff, "%f", &param);
}

bool handle_error(Error<CString> e)
{
	if(e.type == NOT_A_NUMBER || e.type == TOO_SMALL_DISPLAY)
		AfxMessageBox(e.msg, MB_OK | MB_ICONSTOP);	//display error
	else if(e.type == SWAP_VALUES)
	{
		list<CString>::iterator it;
		CString msg = "Warning: The following fields had a maximum that was less than its minimum:\n";
		
		for(it = e.templated_list1.begin(); it != e.templated_list1.end(); ++it)
			msg += *it + "\n";

		msg += "These values have been swapped.  Would you like to continue?";

		if(IDNO == AfxMessageBox(msg, MB_YESNO | MB_ICONEXCLAMATION))
			return false;
		else
			return true;	//continue with code that called this function
	}

	return false;
}