#include "comm_min.h"

//global includes
//#include <set>

//local includes
#include "debug.h"
#include "task_connection.h"
#include "graph.h"

using namespace std;

//externals
extern Graph grp;
extern System sys;

//static globals
static CommMinGraph graph;	//graph used for the communication minimization algorithm

double Do_CommMin_Mapping()
{
/*	multisetConn connections;

	//get all of the source connections from each task
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin();
		conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
			connections.insert(*conn_it);
	}*/

	GraphPartition* best_partition = NULL;

	best_partition = MinCutAlgo();

	string task_name;
	for(int i = 0; i < best_partition->group.size(); ++i)
		for(list<TaskType*>::iterator it = best_partition->group[i].begin(); it != best_partition->group[i].end(); ++it)
			task_name = (*it)->Property()->Get_Name();


/*	multisetTask tasks;
	multisetProc procs;

	double non_allocated_utility = 0;

	//add in all the tasks 
	for(list<TaskType*>::iterator task_init_it = grp.Node_Begin(); task_init_it != grp.Node_End(); ++task_init_it)
		tasks.insert(*task_init_it);

	//add in all of the cpu's
	for(list<Processor*>::iterator proc_init_it = sys.Proc_Begin(); proc_init_it != sys.Proc_End(); ++proc_init_it)
		procs.insert(*proc_init_it);

	//map each task to a processor
	for(multisetTaskIter task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
	{
		list<Processor*> assigned_procs;
		bool assigned_to_proc = false;

		TaskType* task = *task_it;

		list<TaskType*> connected_tasks = Get_Connected_Tasks(task);
		connected_tasks.push_back(task);	//analyze original task also

		//keep trying to assign the connected tasks to a processor.  If it can't be
		//done with the current group remove the least significant task from the group.
		//if all the tasks were removed from the group then the task could not be
		//mapped to a processor - error
		while(connected_tasks.size() && !Assign_Group_To_Proc(connected_tasks, procs))
			connected_tasks = Remove_Task(connected_tasks, task);

		if(connected_tasks.empty())
			non_allocated_utility += task->Property()->Get_Exec_Time() / task->Property()->Get_Period();
	}

	//one or more tasks could not be mapped to a processor
	if(non_allocated_utility)
		return non_allocated_utility;

	//print out the results
	print_CommMin_Mapping();*/

	return 0;
}
/*
list<TaskType*> Get_Connected_Tasks(TaskType* t)
{
	list<TaskType*> connected;

	//get the connections coming in
	for(list<ConnectionType*>::iterator it = t->Get_In_Data_Edges_Begin(); it != t->Get_In_Data_Edges_End(); ++it)
	{
//		if(!(*it)->Get_From()->Property()->Get_Alloc_Proc())	//has a processor already been assigned to this task
			connected.push_back((*it)->Get_From());
	}

	//get the connections going out
	for(it = t->Get_Out_Data_Edges_Begin(); it != t->Get_Out_Data_Edges_End(); ++it)
	{
//		if(!(*it)->Get_From()->Property()->Get_Alloc_Proc())	//has a processor already been assigned to this task
			connected.push_back((*it)->Get_To());
	}

	return connected;
}

double Get_Total_Utilization(list<TaskType*> tasks)
{
	double utilization = 0;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		if(!(*it)->Property()->Get_Alloc_Proc())	//has it already been assigned to a cpu
			utilization += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();
	}

	return utilization;
}

void Put_Tasks_On_Proc(list<TaskType*> tasks, Processor* proc)
{
	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		if(!(*it)->Property()->Get_Alloc_Proc())	//has it already been assigned to a cpu
		{
			(*it)->Property()->Set_Alloc_Proc(proc);
			proc->Add_Task(*it);
		}
	}
}

list<Processor*> Get_Procs_Assigned(list<TaskType*> tasks)
{
	list<Processor*> procs;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		//is there a processor already assigned to the task
		Processor* alloc_proc = (*it)->Property()->Get_Alloc_Proc();
		if(alloc_proc)
		{
			//make sure the processor isn't already in the list
			for(list<Processor*>::iterator proc_it = procs.begin(); proc_it != procs.end(); ++proc_it)
				if(*proc_it == alloc_proc) break;

			if(proc_it == procs.end())
				procs.push_back(alloc_proc);
		}
	}

	return procs;
}

bool Assign_Group_To_Proc(std::list<TaskType*> tasks, multisetProc procs)
{
	multisetProcIter proc_it;

	double utilization = Get_Total_Utilization(tasks);

	list<Processor*> assigned_procs = Get_Procs_Assigned(tasks);

//		processor_most_connected_greater_than gt(connected_tasks, task);
//		assigned_procs.sort(gt);

	//see if the group of connected tasks can be mapped to a processor that has
	//already been assigned
	for(list<Processor*>::iterator assigned_it = assigned_procs.begin(); assigned_it != assigned_procs.end(); ++assigned_it)
	{
		Processor* proc_ptr = *assigned_it;

		//try to put the task and all of its connected tasks onto the processor
		if(proc_ptr->Property()->Allocate_Space(utilization))
		{
			//get the position of proc_ptr in the procs multimap
			for(proc_it = procs.begin(); proc_it != procs.end(); ++proc_it)
				if(*proc_it == proc_ptr) break;

			//if proc_it was not found something is wrong (bug in code)
			if(proc_it != procs.end())
			{
				procs.erase(proc_it);
				Put_Tasks_On_Proc(tasks, proc_ptr);
				procs.insert(proc_ptr);
			}

			return true;
		}
	}

	//the group of connect tasks could not be put on a processor that one of the
	//tasks has already been assigned to, try the rest of the procs
	for(proc_it = procs.begin(); proc_it != procs.end(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		//try to put the task and all of its connected tasks onto the processor
		if(proc_ptr->Property()->Allocate_Space(utilization))
		{
			procs.erase(proc_it);
			Put_Tasks_On_Proc(tasks, proc_ptr);
			procs.insert(proc_ptr);
			return true;
		}
	}

	return false;
}

std::list<TaskType*> Remove_Task(std::list<TaskType*> tasks, TaskType* root_task)
{
	tasks.pop_back();

	return tasks;
}

//function object definitions
bool connection_greater_than::operator () (ConnectionType* c1, ConnectionType* c2)
{
	double bytes_per_second1;
	double bytes_per_second2;

	bytes_per_second1 = c1->Property()->Get_Message_Size() * c1->Property()->Get_Frequency();
	bytes_per_second2 = c2->Property()->Get_Message_Size() * c2->Property()->Get_Frequency();

	return bytes_per_second1 > bytes_per_second2;
}

bool task_connection_size_greater_than::operator () (TaskType* t1, TaskType* t2)
{
	return Get_Total_Connection_Size(t1) > Get_Total_Connection_Size(t2);
}

double Get_Total_Connection_Size(TaskType* t)
{
	double bytes_per_second = 0;

	//in connections
	for(list<ConnectionType*>::iterator it = t->Get_In_Data_Edges_Begin(); it != t->Get_In_Data_Edges_End(); ++it)
		bytes_per_second += (*it)->Property()->Get_Message_Size() * (*it)->Property()->Get_Frequency();

	//out connections
	for(it = t->Get_Out_Data_Edges_Begin(); it != t->Get_Out_Data_Edges_End(); ++it)
		bytes_per_second += (*it)->Property()->Get_Message_Size() * (*it)->Property()->Get_Frequency();
	
	return bytes_per_second;
}

bool processor_most_connected_greater_than::operator () (Processor* p1, Processor* p2)
{
	return greater_than_helper(p1) > greater_than_helper(p2);
}

double processor_most_connected_greater_than::greater_than_helper(Processor* p)
{
	double bytes_per_second;

	ConnectionType* connection;

	for(list<TaskType*>::iterator task_it = p->Tasks_Begin(); task_it != p->Tasks_End(); ++task_it)
	{
		TaskType* task_ptr = *task_it;

		if(task_ptr == root_task)		//don't try to get the connection from the root task to itself
			continue;

		//make sure this task is in the task list given
		for(list<TaskType*>::iterator check_it = task_list.begin(); check_it != task_list.end(); ++check_it)
			if(*check_it == task_ptr) break;

		if(check_it == task_list.end())	//not in the list given
			continue;

		connection = root_task->Get_Connection_From(task_ptr);

		//connection was not from task_ptr to root_task try from root_task to task_ptr
		if(!connection)										
			connection = root_task->Get_Connection_To(task_ptr);

		//root_task and task_ptr are not connected - given list is not correct
		if(!connection)
			continue;

		bytes_per_second += connection->Property()->Get_Message_Size() * connection->Property()->Get_Frequency();
	}

	return bytes_per_second;
}*/

GraphPartition* MinCutAlgo()
{
	GraphPartition*	best_partition = new GraphPartition;
	GraphPartition last_partition;

	best_partition->group.resize(sys.Get_Number_Of_Procs());
	last_partition.group.resize(sys.Get_Number_Of_Procs());
	
	PopulateGraph();

//	best_partition = MinimumCut();

	MinimumCut(&last_partition, graph.GetNodes());

	if(&last_partition)
		*best_partition = last_partition;

	if(best_partition->cost == 0)
		return best_partition;

	while(graph.GetEdges().size() >= sys.Get_Number_Of_Procs())
	{
		MinimumCut(&last_partition, graph.GetNodes());

		if(&last_partition && last_partition.cost < best_partition->cost)
			*best_partition = last_partition;

		if(best_partition->cost == 0)
			return best_partition;
	}

	//find the best partition
/*	while(graph.GetEdges().size() > 1)
	{
		last_partition = MinimumCut();

		if(last_partition->cost == 0)
		{
			if(best_partition)
				delete best_partition;

			return last_partition;
		}

		if(!best_partition)
			best_partition = last_partition;
		else if(last_partition && last_partition->cost < best_partition->cost)
		{
			delete best_partition;
			best_partition = last_partition;
		}
		else if(last_partition)
			delete last_partition;
	}*/

	return best_partition;
}

void PopulateGraph()
{
	//add in all of the nodes - needs to be done before adding edges
	for(list<TaskType*>::iterator task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		CommMinNode* new_task = new CommMinNode(*task_it);
		graph.AddNode(new_task);
	}

	//add in all of the edges
	for(task_it = grp.Node_Begin(); task_it != grp.Node_End(); ++task_it)
	{
		//add in connections that are coming out of this connection
		for(list<ConnectionType*>::iterator conn_it = (*task_it)->Get_Out_Data_Edges_Begin(); 
		conn_it != (*task_it)->Get_Out_Data_Edges_End(); ++conn_it)
		{
			TaskType* from = (*conn_it)->Get_From();
			TaskType* to = (*conn_it)->Get_To();

			CommMinNode* node1 = graph.FindNodeFromTask(from);
			CommMinNode* node2 = graph.FindNodeFromTask(to);

			double bps = (*conn_it)->Property()->Get_Frequency() * (*conn_it)->Property()->Get_Message_Size();

			CommMinEdge* new_edge = new CommMinEdge(bps, node1, node2);

			//add the new edge to its nodes and to the graph
			node1->AddEdge(new_edge);
			node2->AddEdge(new_edge);
			graph.AddEdge(new_edge);
		}
	}
}

GraphPartition* MinimumCut()
{
	GraphPartition* partition = new GraphPartition;
	CommMinNode* s = NULL;
	CommMinNode* t = NULL;
	int last_group = 0;
	int curr_group = 0;

	list<pair<CommMinNode*, CommMinNode*> > merged_nodes;

	partition->group.resize(sys.Get_Number_Of_Procs());

	bool all_tasks_mapped = true;

	list<CommMinNode*> nodes = graph.GetNodes();

	//initialize the first group with first node in the list
	CommMinNode* np = nodes.front();

	list<Processor*>::iterator proc_it = sys.Proc_Begin();
	for(int i = 0; i < partition->group.size(); ++i)
	{	
		if(np->GetTotalUtilization() <= (*proc_it)->Property()->Get_Bound())
		{
			partition->AddTasks(i, np);
			last_group = i;
			break;
		}

		proc_it++;
	}
	if(proc_it == sys.Proc_End())
		all_tasks_mapped = false;

	nodes.pop_front();

	while(!nodes.empty())
	{
		list<Processor*>::iterator proc_it = sys.Proc_Begin();
		for(int i = last_group; i < partition->group.size(); ++i)
		{
			t = FindBestNodeNew(i, partition, nodes);	//find the best node to add to the partition
		
			if(partition->GetTotalUtilization(i) + t->GetTotalUtilization() <= (*proc_it)->Property()->Get_Bound())
			{
				partition->AddTasks(i, t);
				curr_group = i;
				break;
			}

			proc_it++;
		}

		if(proc_it == sys.Proc_End())
			all_tasks_mapped = false;
		else
		{
			if(curr_group != last_group && s)
			{
				//merge s and t
				t->Merge(s);

				graph.RemoveNode(s);
				graph.RemoveEdgesFrom(s);

				delete s;
			}

			last_group = curr_group;
			s = t; 
		}

		nodes.remove(s);
		
/*		if(nodes.size() == 1)
		{
			proc_it = sys.Proc_Begin();
			for(int i = last_group; i < partition->group.size(); ++i)
			{
				t = nodes.front();
				if(partition->GetTotalUtilization(i) + t->GetTotalUtilization() <= (*proc_it)->Property()->Get_Bound())
				{
					partition->AddTasks(i, t);
					break;
				}

				proc_it++;
			}

			if(proc_it == sys.Proc_End())
				all_tasks_mapped = false;

			nodes.pop_front();
		}*/
	}

	//find the cost of splitting these to groups up
	partition->cost = findPartitionCost(partition);

	//s and t should always be non NULL - this if statement is an added precaution
/*	if(s && t)
	{
		//merge s and t together and remove t and its edges from the graph
		s->Merge(t);
		graph.RemoveNode(t);
		graph.RemoveEdgesFrom(t);

		delete t;
	}*/

/*	for(list<pair<CommMinNode*,CommMinNode*> >::iterator merge_it = merged_nodes.begin();
	merge_it != merged_nodes.end(); ++merge_it)
	{
		merge_it->first->Merge(merge_it->second);
		graph.RemoveNode(merge_it->second);
		graph.RemoveEdgesFrom(merge_it->second);

		delete merge_it->second;
	}*/

	if(all_tasks_mapped)
		return partition;
	else
	{
		delete partition;
		return NULL;
	}
}

CommMinNode* FindBestNode(GraphPartition partition, list<CommMinNode*> nodes)
{
	CommMinNode* best;
	double saved_cost;
	double best_saved_cost;

	best			= nodes.front();					//initialize the best node as the first node
	best_saved_cost	= find_saved_cost(best, partition); //find the saved cost of adding it to the partition

	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
	{
		saved_cost = find_saved_cost(*it, partition);

		//check to see if this node saves more then the previous best node
		if(saved_cost > best_saved_cost)
		{
			best_saved_cost = saved_cost;
			best = *it;
		}
	}

	return best;
}

CommMinNode* FindBestNodeNew(int group_num, GraphPartition* partition, list<CommMinNode*> nodes)
{
	CommMinNode* best;
	double saved_cost;
	double best_saved_cost;

	best			= nodes.front();					//initialize the best node as the first node
	best_saved_cost	= findSavedCost(group_num, best, partition); //find the saved cost of adding it to the partition

	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
	{
		saved_cost = findSavedCost(group_num, *it, partition);

		//check to see if this node saves more then the previous best node
		if(saved_cost > best_saved_cost)
		{
			best_saved_cost = saved_cost;
			best = *it;
		}
	}

	return best;
}

double find_saved_cost(CommMinNode* node, GraphPartition partition)
{
	double saved_cost = 0;
	list<TaskType*> tasks = partition.group1;
	list<CommMinNode*> checked_nodes;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		CommMinNode* it_node = graph.FindNodeFromTask(*it);

		//check if this node has already been tested, if so don't check it again
		if(onList(it_node, checked_nodes))
			continue;
		checked_nodes.push_back(it_node);

		list<CommMinEdge*> edges = it_node->GetEdges();

		//go through each of the edges in the partition and see if they are connected 
		//to the given node - if so add in the edge's bps to the saved cost
		for(list<CommMinEdge*>::iterator conn_it = it_node->GetEdgeBegin(); conn_it != it_node->GetEdgeEnd(); ++conn_it)
		{
			CommMinEdge* edge = *conn_it;

			CommMinNode* first_node = edge->GetFirstConnectedNode();
			CommMinNode* second_node = edge->GetSecondConnectedNode();

			if(first_node == node || second_node == node)
				saved_cost += edge->GetMessageSize();
		}
	}

	return saved_cost;
}

double findSavedCost(int group_num, CommMinNode* node, GraphPartition* partition)
{
	double saved_cost = 0;
	list<TaskType*> tasks = partition->group[group_num];
	list<CommMinNode*> checked_nodes;

	for(list<TaskType*>::iterator it = tasks.begin(); it != tasks.end(); ++it)
	{
		CommMinNode* it_node = graph.FindNodeFromTask(*it);

		//check if this node has already been tested, if so don't check it again
		if(onList(it_node, checked_nodes))
			continue;
		checked_nodes.push_back(it_node);

		list<CommMinEdge*> edges = it_node->GetEdges();

		//go through each of the edges in the partition and see if they are connected 
		//to the given node - if so add in the edge's bps to the saved cost
		for(list<CommMinEdge*>::iterator conn_it = it_node->GetEdgeBegin(); conn_it != it_node->GetEdgeEnd(); ++conn_it)
		{
			CommMinEdge* edge = *conn_it;

			CommMinNode* first_node = edge->GetFirstConnectedNode();
			CommMinNode* second_node = edge->GetSecondConnectedNode();

			if(first_node == node || second_node == node)
				saved_cost += edge->GetMessageSize();
		}
	}

	return saved_cost;
}

double findPartitionCost(GraphPartition* partition)
{
	double cost = 0;
	double edge_cost;

	CommMinNode* node1;
	CommMinNode* node2;

	list<CommMinEdge*> edges = graph.GetEdges();

	//go through each edge and determine if its nodes are in different groups - if so
	//then add the edge cost to the total cost
	for(list<CommMinEdge*>::iterator edge_it = edges.begin(); edge_it != edges.end(); ++edge_it)
	{
		int node1group;
		int node2group;

		node1		= (*edge_it)->GetFirstConnectedNode();
		node2		= (*edge_it)->GetSecondConnectedNode();
		edge_cost   = (*edge_it)->GetMessageSize();

		node1group  = getGroupNew(node1, partition);
		node2group  = getGroupNew(node2, partition);

		if(node1group != node2group)
			cost += edge_cost;
	}

	return cost;
}

bool onList(CommMinNode* node_check, list<CommMinNode*>  checked_nodes)
{
	for(list<CommMinNode*>::iterator it = checked_nodes.begin(); it != checked_nodes.end(); ++it)
		if(node_check == *it) return true;

	return false;
}

int getGroup(CommMinNode* node, GraphPartition partition)
{
	list<CommMinNode*> checked_nodes;
	list<TaskType*> tasks = partition.group2;

	for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
	{
		CommMinNode* node_check = graph.FindNodeFromTask(*task_it);

		//make sure that the node has not already been previously checked
		if(onList(node_check, checked_nodes))
			continue;
		else
			checked_nodes.push_back(node_check);
		
		//found node
		if(node_check == node) break;
	}

	if(task_it == tasks.end())	//node not found in group2 - must be in group1
		return 1;
	else						//node found in group2
		return 2;
}

int getGroupNew(CommMinNode* node, GraphPartition* partition)
{
	list<CommMinNode*> checked_nodes;

	for(int i = 0; i < partition->group.size(); ++i)
	{
		list<TaskType*> tasks = partition->group[i];
		for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
		{
			CommMinNode* node_check = graph.FindNodeFromTask(*task_it);

			//make sure that the node has not already been previously checked
			if(onList(node_check, checked_nodes))
				continue;
			else
				checked_nodes.push_back(node_check);
			
			//found node
			if(node_check == node)
				return i;
		}
	}

	//could not find the node
	return -1;
}

/****************** CommMinNode Definitions *********************/

CommMinNode::CommMinNode(TaskType* task)
{
	tasks.push_back(task);
	total_utilization = task->Property()->Get_Exec_Time() / task->Property()->Get_Period();
}

CommMinNode::~CommMinNode()
{
/*	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
	{
		CommMinEdge* edge = *it;
		delete edge;
	}*/
}

bool CommMinNode::operator == (CommMinNode n)
{
	//make sure it has all of the same tasks
	for(list<TaskType*>::iterator i_it = tasks.begin(); i_it != tasks.end(); ++i_it)
	{
		for(list<TaskType*>::iterator j_it = n.GetTasksBegin(); j_it != n.GetTasksEnd(); ++j_it)
			if(*j_it == *i_it) break;

		if(j_it == n.GetTasksEnd())
			return false;
	}

	//make sure it has all of the same edges
	for(list<CommMinEdge*>::iterator e_it = edges.begin(); e_it != edges.end(); ++e_it)
	{
		for(list<CommMinEdge*>::iterator j_it = n.GetEdgeBegin(); j_it != n.GetEdgeEnd(); ++j_it)
			if(*j_it == *e_it) break;

		if(j_it == n.GetEdgeEnd())
			return false;
	}

	return true;
}

bool CommMinNode::Merge(CommMinNode* node)
{
	list<CommMinEdge*>	edge_list = node->GetEdges();
	list<TaskType*>		task_list = node->GetTasks();

	//check to see if the nodes are connected
	for(list<CommMinEdge*>::iterator edge_it = edge_list.begin(); edge_it != edge_list.end(); ++edge_it)
	{
		if((*edge_it)->GetFirstConnectedNode() == this || (*edge_it)->GetSecondConnectedNode() == this)
			break;
	}

	//nodes are not connected, cannot merge
	if(edge_it == edges.end())
		return false;

	//add in the tasks from the node to be merged
	for(list<TaskType*>::iterator task_it = task_list.begin(); task_it != task_list.end(); ++task_it)
		tasks.push_back(*task_it);

	//update the total utilization
	total_utilization += node->GetTotalUtilization();

	//add all edges from the node to be merged or add values to existing edges
	for(edge_it = edge_list.begin(); edge_it != edge_list.end(); ++edge_it)
	{
		CommMinEdge* edge_ptr = *edge_it;
		CommMinNode* check_node;

		//find node that is connected to the node to be merged
		if(edge_ptr->GetFirstConnectedNode() == node)
			check_node = edge_ptr->GetSecondConnectedNode();
		else if(edge_ptr->GetSecondConnectedNode() == node)
			check_node = edge_ptr->GetFirstConnectedNode();
		else
			return false;	//node to be merged was not part of this connection - error

		//ignore the connection from the node to be merged to this node
		if(check_node == this)
			continue;

		//check to see if the edge is currently part of this node
		for(list<CommMinEdge*>::iterator cur_edge_it = edges.begin(); cur_edge_it != edges.end(); ++cur_edge_it)
		{
			//edge found add in its weight
			if((*cur_edge_it)->GetFirstConnectedNode() == check_node || 
			   (*cur_edge_it)->GetSecondConnectedNode() == check_node)	
			{
				(*cur_edge_it)->AddMessageSize((*edge_it)->GetMessageSize());
				break;
			}
		}

		//edge was not currently part of this node - add it in
		if(cur_edge_it == edges.end())
		{
			//move the connection from the merged node to this node
			if(edge_ptr->GetFirstConnectedNode() == node)
				edge_ptr->SetFirstConnectedNode(this);
			else if(edge_ptr->GetSecondConnectedNode() == node)
				edge_ptr->SetSecondConnectedNode(this);
	
			edges.push_back(edge_ptr);	//add edge to this node
			node->RemoveEdge(edge_ptr);	//remove edge from old node
		}
	}

	return true;
}

void CommMinNode::RemoveEdge(CommMinEdge* edge)
{
	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
		if(edge == *it)
		{
			edges.erase(it);
			return;
		}
}

/*______________________________________________________________*/


/****************** CommMinEdge Definitions *********************/

CommMinEdge::CommMinEdge(double bytes_per_second, CommMinNode* first_node, CommMinNode* second_node)
{
	 bps   = bytes_per_second;
	 node1 = first_node;
	 node2 = second_node; 
}

/*______________________________________________________________*/


/******************* CommMinGraph Definitions *******************/

CommMinNode* CommMinGraph::FindNodeFromTask(TaskType* task)
{
	for(list<CommMinNode*>::iterator node_it = nodes.begin(); node_it != nodes.end(); ++node_it)
	{
		list<TaskType*> tasks = (*node_it)->GetTasks();

		for(list<TaskType*>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
			if(*task_it == task) return *node_it;
	}

	return NULL;
}

void CommMinGraph::RemoveEdgesFrom(CommMinNode* t)
{
	for(list<CommMinEdge*>::iterator it = t->GetEdgeBegin(); it != t->GetEdgeEnd(); ++it)
	{
		CommMinEdge* edge = *it;

		//remove the edge from the node that is not part of t
		if(edge->GetFirstConnectedNode() != t)
			edge->GetFirstConnectedNode()->RemoveEdge(edge);
		else
			edge->GetSecondConnectedNode()->RemoveEdge(edge);

		//remove the edge from the graph
		RemoveEdge(edge);

		delete edge;
	}
}

void CommMinGraph::RemoveNode(CommMinNode* node)
{
	for(list<CommMinNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it)
		if(*it == node)
		{
			nodes.erase(it);
			return;
		}
}

void CommMinGraph::RemoveEdge(CommMinEdge* edge)
{
	for(list<CommMinEdge*>::iterator it = edges.begin(); it != edges.end(); ++it)
		if(*it == edge)
		{
			edges.erase(it);
			return;
		}
}

/*______________________________________________________________*/


/******************* Graph Partition Defintions *****************/

GraphPartition::GraphPartition(const GraphPartition& g)
{
	cost = g.cost;
	group1 = g.group1;
	group2 = g.group2;
}

GraphPartition& GraphPartition::operator = (const GraphPartition& g)
{
	if(this != &g)
	{
		cost = g.cost;
		group1 = g.group1;
		group2 = g.group2;
	}

	return *this;
}

void GraphPartition::AddTasksToFirst(CommMinNode* node)
{
	for(list<TaskType*>::iterator it = node->GetTasksBegin(); it != node->GetTasksEnd(); ++it)
		group1.push_back(*it);
}

void GraphPartition::AddTasksToSecond(CommMinNode* node)
{
	for(list<TaskType*>::iterator it = node->GetTasksBegin(); it != node->GetTasksEnd(); ++it)
		group2.push_back(*it);
}

void GraphPartition::AddTasks(int group_num, CommMinNode* node)
{
	for(list<TaskType*>::iterator it = node->GetTasksBegin(); it != node->GetTasksEnd(); ++it)
		group[group_num].push_back(*it);
}

double GraphPartition::GetFirstTotalUtilization()
{
	double utilization = 0;

	for(list<TaskType*>::iterator it = group1.begin(); it != group1.end(); ++it)
		utilization += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return utilization;
}

double GraphPartition::GetSecondTotalUtilization()
{
	double utilization = 0;

	for(list<TaskType*>::iterator it = group2.begin(); it != group2.end(); ++it)
		utilization += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return utilization;
}

double GraphPartition::GetTotalUtilization(int group_num)
{
	double utilization = 0;

	for(list<TaskType*>::iterator it = group[group_num].begin(); it != group[group_num].end(); ++it)
		utilization += (*it)->Property()->Get_Exec_Time() / (*it)->Property()->Get_Period();

	return utilization;
}
/*______________________________________________________________*/






void show_partition(GraphPartition* partition)
{
	string name;
	int group;

	for(group = 0; group < partition->group.size(); ++group)
		for(list<TaskType*>::iterator it = partition->group[group].begin(); it != partition->group[group].end(); ++it)
			name = (*it)->Property()->Get_Name();
}

void MinimumCut(GraphPartition* partition, list<CommMinNode*> nodes)
{
	CommMinNode* s = NULL;
	CommMinNode* t = NULL;
	GraphPartition given_partition = *partition;
	list<CommMinNode> grouped_nodes;
	double cost;

	if(valid_partition(partition))
		return;


	CommMinNode* np = nodes.front();
	partition->AddTasks(1, np);
	nodes.pop_front();

	while(!nodes.empty())
	{
		s = FindBestNodeNew(1, partition, nodes);	//find the best node to add to the partition
		partition->AddTasks(1, s);
		nodes.remove(s);
		grouped_nodes.push_back(*s);

		if(nodes.size() == 1)
		{
			t = nodes.front();
			partition->AddTasks(0, t);
			nodes.pop_front();
		}
	} 

	//find the cost of splitting these to groups up
	partition->cost = findPartitionCost(partition);

	if(partition->group.size() > 2)
	{
		MinimumCutRecursive(partition, grouped_nodes);
		cost = findPartitionCost(partition);

		if(partition->cost < cost)
			*partition = given_partition;
		else
			partition->cost = cost;
	}

	//s and t should always be non NULL - this if statement is just an added precaution
	if(s && t)
	{
		//merge s and t together and remove t and its edges from the graph
		s->Merge(t);
		graph.RemoveNode(t);
		graph.RemoveEdgesFrom(t);

		delete t;
	}

} //MinimumCut

bool valid_partition(GraphPartition* partition)
{
	return false;
}

void MinimumCutRecursive(GraphPartition* partition, std::list<CommMinNode> nodes)
{
	int group_num;
	CommMinNode s;
	CommMinNode t;
	GraphPartition given_partition = *partition;
	list<CommMinNode> grouped_nodes;
	double cost;

	if(valid_partition(partition))
		return;

	for(group_num = 0; group_num < partition->group.size(); ++group_num)
		if(partition->group[group_num].empty()) break;

	partition->group[--group_num].clear();

	partition->AddTasks(group_num + 1, &(nodes.front()));
	nodes.pop_front();

	while(!nodes.empty())
	{
		s = FindBestNodeRecurs(group_num + 1, partition, nodes);	//find the best node to add to the partition
		partition->AddTasks(group_num + 1, &s);
		nodes.remove(s);
		grouped_nodes.push_back(s);

		if(nodes.size() == 1)
		{
			t = nodes.front();
			partition->AddTasks(group_num, &t);
			nodes.pop_front();
		}
	} 

	partition->cost = findPartitionCost(partition);

	if(partition->group[partition->group.size()-1].empty())
	{
		MinimumCutRecursive(partition, grouped_nodes);
		cost = findPartitionCost(partition);

		if(partition->cost < cost)
			*partition = given_partition;
		else
			partition->cost = cost;
	}
}

CommMinNode FindBestNodeRecurs(int group_num, GraphPartition* partition, list<CommMinNode> nodes)
{
	CommMinNode best;
	double saved_cost;
	double best_saved_cost;

	best			= nodes.front();					//initialize the best node as the first node
	best_saved_cost	= findSavedCost(group_num, &best, partition); //find the saved cost of adding it to the partition

	for(list<CommMinNode>::iterator it = nodes.begin(); it != nodes.end(); ++it)
	{
		saved_cost = findSavedCost(group_num, &(*it), partition);

		//check to see if this node saves more then the previous best node
		if(saved_cost > best_saved_cost)
		{
			best_saved_cost = saved_cost;
			best = *it;
		}
	}

	return best;
}
