#include "..\stdafx.h"
#include "windows.h"
#include <cstdio>
#include <algorithm>
#include "graph.h"
#include "table.h"
#include "graphalgo.h"

typedef struct
{
    TaskType *node;
    bool visited;
} bfs_tmp_data;

static float calculate_edge_weight(TaskType *from, ConnectionType *edge);
static list<sssp_entry>::iterator get_maximum (list<sssp_entry>&);
static list<sssp_entry>::iterator find_iterator (list<sssp_entry>&, TaskType *ptask);
static void relax (list<sssp_entry>::iterator&, list<sssp_entry>::iterator&);
static void set_bfs_mark (TaskType *, bool, list<bfs_tmp_data>&);
static bool get_bfs_mark (TaskType *, list<bfs_tmp_data>&);
static void topo_sort_tmp (TaskType *, list<bfs_tmp_data>&,
        list<TaskType *>*);
static void priority_topo_sort_tmp (TaskType *, list<bfs_tmp_data>&,
        list<TaskType *>*);
static TaskType* get_min_priority (list<TaskType*>& l);

list<sssp_entry>* SSSP_Dijkstra (Graph& g, TaskType& s)
{
    sssp_entry sse;
    list<sssp_entry> *sl, sq;
    list<TaskType *>::iterator si;

    sl=new list<sssp_entry>;

    for(si=g.Node_Begin();si!=g.Node_End();si++)
    {
        sse.ptask=*si;
        sse.distance=NIL;
        sse.pathlength=NIL;
        sse.pinedge=0;
        if(sse.ptask==&s) 
        {
            sse.distance=(*si)->Property()->Get_Exec_Time();
            sse.pathlength=1;
        }
        sq.push_front(sse);
    }
    
    while(!sq.empty())
    {
        list<sssp_entry>::iterator smax=get_maximum(sq);
    
        sl->push_front(*smax);

        TaskType *spt=(*smax).ptask;
        list<ConnectionType *>::iterator siv;
        
        for(siv=spt->Get_Out_Edges_Begin();siv!=spt->Get_Out_Edges_End();siv++)
        {
            list<sssp_entry>::iterator snu=find_iterator(sq, (*siv)->Get_To());

            if(snu!=sq.end()) relax(smax, snu);
        }           
        
        sq.erase(smax);        
    }

    return sl;
}

list<sssp_entry>* SSSP_Dijkstra (Graph& g)
{
    sssp_entry sse;
    list<sssp_entry> *sl, sq;
    list<TaskType *>::iterator si;

    sl=new list<sssp_entry>;

    for(si=g.Node_Begin();si!=g.Node_End();si++)
    {
        sse.ptask=*si;
        sse.distance=NIL;
        sse.pathlength=NIL;
        sse.pinedge=0;
        if(!(sse.ptask->Get_Number_Of_In_Connection())) 
        {
            sse.distance=(*si)->Property()->Get_Exec_Time();
            sse.pathlength=1;
        }
        sq.push_front(sse);
    }
    
    while(!sq.empty())
    {
        list<sssp_entry>::iterator smax=get_maximum(sq);
    
        sl->push_front(*smax);

        TaskType *spt=(*smax).ptask;
        list<ConnectionType *>::iterator siv;
        
        for(siv=spt->Get_Out_Edges_Begin();siv!=spt->Get_Out_Edges_End();siv++)
        {
            list<sssp_entry>::iterator snu=find_iterator(sq, (*siv)->Get_To());

            if(snu!=sq.end()) relax(smax, snu);
        }           
        
        sq.erase(smax);        
    }

    return sl;
}

list<sssp_entry>* SSSP_Bellman_Ford (Graph& g, TaskType& s)
{
    sssp_entry sse;
    list<sssp_entry> *sl;
    list<TaskType *>::iterator si;
    int i;
    
    sl=new list<sssp_entry>;

    for(si=g.Node_Begin();si!=g.Node_End();si++)
    {
        sse.ptask=*si;
        sse.distance=NIL;
        sse.pathlength=NIL;
        sse.pinedge=0;
        if(sse.ptask==&s) 
        {
            sse.distance=(*si)->Property()->Get_Exec_Time();
            sse.pathlength=1;
        }
        sl->push_front(sse);
    }
    
    for(i=0;i<g.Get_Number_Of_Nodes()-1;i++)
    {
        list<ConnectionType *>::iterator sic;

        for(sic=g.Edge_Begin();sic!=g.Edge_End();sic++)
        {
            list<sssp_entry>::iterator u=find_iterator(*sl, (*sic)->Get_From());
            list<sssp_entry>::iterator v=find_iterator(*sl, (*sic)->Get_To());
            
            relax(u,v);
        }
    }

    return sl;
}

list<sssp_entry>* SSSP_Bellman_Ford (Graph& g)
{
    sssp_entry sse;
    list<sssp_entry> *sl;
    list<TaskType *>::iterator si;
    int i;
    
    sl=new list<sssp_entry>;

    for(si=g.Node_Begin();si!=g.Node_End();si++)
    {
        sse.ptask=*si;
        sse.distance=NIL;
        sse.pathlength=NIL;
        sse.pinedge=0;
        if(!(sse.ptask->Get_Number_Of_In_Connection())) 
        {
            sse.distance=(*si)->Property()->Get_Exec_Time();
            sse.pathlength=1;
        }
        sl->push_front(sse);
    }
    
    for(i=0;i<g.Get_Number_Of_Nodes()-1;i++)
    {
        list<ConnectionType *>::iterator sic;

        for(sic=g.Edge_Begin();sic!=g.Edge_End();sic++)
        {
            list<sssp_entry>::iterator u=find_iterator(*sl, (*sic)->Get_From());
            list<sssp_entry>::iterator v=find_iterator(*sl, (*sic)->Get_To());
            
            relax(u,v);
        }
    }

    return sl;
}

// BFS Traverse and Apply func

void BFS_Traversal (Graph& g, TaskType *t, void (*func) (TaskType *, TaskType *, ConnectionType *))
{

    list<bfs_tmp_data> btdl;
    list<TaskType *>::iterator bi;

    for(bi=g.Node_Begin();bi!=g.Node_End();bi++)
    {
        bfs_tmp_data btd;

        btd.node=*bi;
        btd.visited=false;
        btdl.push_front(btd);
    }

    list<TaskType *> bfsq;
    bfsq.push_back(t);
    while(!bfsq.empty())
    {
        TaskType * bt=bfsq.front();
        bfsq.pop_front();

        set_bfs_mark(bt,true, btdl);  // Set bt as visited
        list<ConnectionType *>::iterator bci;

        for(bci=bt->Get_Out_Edges_Begin();bci!=bt->Get_Out_Edges_End();bci++)
        {
            ConnectionType *bc=*bci;

            if(!get_bfs_mark(bc->Get_To(), btdl))
            {
                if(func) func(bt, bc->Get_To(), bc);
                bfsq.push_back(bc->Get_To());
            }
        }
    }
    return;
}

list<TaskType *>* Topo_Sort (Graph& g)
{
    list<bfs_tmp_data> tbtd;
    list<TaskType *>* sorted = new list<TaskType *>;
    list<TaskType *> dfsq;
    list<TaskType *>::iterator ti;

    for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
    {
        bfs_tmp_data btd;
        
        btd.node=*ti;
        btd.visited=false;
        tbtd.push_front(btd);
    }

    for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
    {
        TaskType *tv=*ti;
        
        if(!get_bfs_mark(tv, tbtd)) topo_sort_tmp(tv,tbtd,sorted);
    }
    return sorted;
}

void Topo_Sort (list<TaskType*>& sorted, TaskType* root, list<TaskType*>& g)
{
    list<bfs_tmp_data> tbtd;
    list<TaskType *> dfsq;
    list<TaskType *>::iterator ti;

    for(ti=g.begin();ti!=g.end();ti++)
    {
        bfs_tmp_data btd;
        
        if((*ti)->Property()->Get_Global_Root()==root)
        {
            btd.node=*ti;
            btd.visited=false;
            tbtd.push_front(btd);
        }
    }

    topo_sort_tmp(root,tbtd,&sorted);
}

void Topo_Sort (list<TaskType*>& sorted, TaskType* root, Graph& g)
{
    list<bfs_tmp_data> tbtd;
    list<TaskType *> dfsq;
    list<TaskType *>::iterator ti;

    for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
    {
        bfs_tmp_data btd;
        
        if((*ti)->Property()->Get_Global_Root()==root)
        {
            btd.node=*ti;
            btd.visited=false;
            tbtd.push_front(btd);
        }
    }

    topo_sort_tmp(root,tbtd,&sorted);
}

// Same as Topo Sort, but uses priority between concurrent tasks

void Priority_Topo_Sort (list<TaskType*>& sorted, TaskType* root, Graph& g)
{
    list<bfs_tmp_data> tbtd;
    list<TaskType *> dfsq;
    list<TaskType *>::iterator ti;

    for(ti=g.Node_Begin();ti!=g.Node_End();ti++)
    {
        bfs_tmp_data btd;
        
        if((*ti)->Property()->Get_Global_Root()==root)
        {
            btd.node=*ti;
            btd.visited=false;
            tbtd.push_front(btd);
        }
    }

    priority_topo_sort_tmp(root,tbtd,&sorted);
}

void get_pred_task_list(list<TaskType*>& ptl, TaskType* pt)
{
    list<ConnectionType*>::iterator gci;
            
    if(pt->Get_Number_Of_In_Connection()==0) return;
    for(gci=pt->Get_In_Edges_Begin();gci!=pt->Get_In_Edges_End();gci++)
    {
        ConnectionType* pc=*gci;
        TaskType* pp=pc->Get_From();
                                               
        ptl.push_front(pp);
        get_pred_task_list(ptl,pp);
    }
    return;
}

void get_succ_task_list(list<TaskType*>& ptl, TaskType* pt)
{
    list<ConnectionType*>::iterator gci;
            
    if(pt->Get_Number_Of_Out_Connection()==0) return;
    for(gci=pt->Get_Out_Edges_Begin();gci!=pt->Get_Out_Edges_End();gci++)
    {
        ConnectionType* pc=*gci;
        TaskType* pp=pc->Get_To();
                                               
        list<TaskType*>::iterator si=find(ptl.begin(),ptl.end(),pp);
        if(si!=ptl.end()) continue;
        ptl.push_front(pp);
        get_succ_task_list(ptl,pp);
    }
    return;
}

void get_control_succ_task_list(list<TaskType*>& ptl, TaskType* pt)
{
    list<ConnectionType*>::iterator gci;
            
    if(pt->Get_Number_Of_Out_Connection()==0) return;
    for(gci=pt->Get_Out_Edges_Begin();gci!=pt->Get_Out_Edges_End();gci++)
    {
        ConnectionType* pc=*gci;

        if(!(pc->Property()->Is_Control_Flow())) continue;
        TaskType* pp=pc->Get_To();
                                               
        list<TaskType*>::iterator si=find(ptl.begin(),ptl.end(),pp);
        if(si!=ptl.end()) continue;
        ptl.push_front(pp);
        get_control_succ_task_list(ptl,pp);
    }
    return;
}

void get_sub_graph(Graph& g, list<TaskType*>& tl)
{
    list<TaskType*>::iterator gti;

    g.Clear();
    for(gti=tl.begin();gti!=tl.end();gti++)
    {
        list<ConnectionType*>::iterator gci;
        TaskType *pt=*gti;
        
        g.Add_Node(pt);
        for(gci=pt->Get_Out_Edges_Begin();gci!=pt->Get_Out_Edges_End();gci++)
        {
            ConnectionType *pc=*gci;

            g.Add_Edge(pc);
        }
    }
}

void set_dependency_table(Table<TaskType*,bool>& dep_table, Graph& g)
{
    list<TaskType*>::iterator it;
    
    dep_table.Clear();
    dep_table.Set_Number_Of_Elements(g.Get_Number_Of_Nodes());
    
    for(it=g.Node_Begin();it!=g.Node_End();it++)
    {
        TaskType* pt=*it;

        dep_table.Add_Element(pt);
    }
    
    dep_table.Initialize(false);
    
    for(it=g.Node_Begin();it!=g.Node_End();it++)
    {
        TaskType* pt=*it;
        list<TaskType*> succ_list;
        list<TaskType*>::iterator isc;

        get_control_succ_task_list(succ_list,pt);

        for(isc=succ_list.begin();isc!=succ_list.end();isc++)
        {
            TaskType* pt2=*isc;

            dep_table.Set_Value(pt,pt2,true);
        }
    }
    return;
}

// Get parent task in the SSSP entry

TaskType* get_parent_task(TaskType *pt, list<sssp_entry>* pl)
{
    list<sssp_entry>::iterator gi;

    for(gi=pl->begin();gi!=pl->end();gi++)
    {
        sssp_entry gse=*gi;

        if(gse.ptask==pt) 
            return (gse.pinedge)?gse.pinedge->Get_From():0;
    }
	return NULL;
}

// Calculate Edge weights

static float calculate_edge_weight(TaskType *from, ConnectionType *edge)
{
    return (from->Property()->Get_Exec_Time()+edge->Property()->Get_Trans_Time());
}

// Relaxing operation

static void relax (list<sssp_entry>::iterator& u, list<sssp_entry>::iterator& v)
{
    ConnectionType *pcn=(*v).ptask->Get_Connection_From ((*u).ptask);
    float w=calculate_edge_weight((*v).ptask, pcn);
    if((*v).distance < (*u).distance + w)
    {
        (*v).distance=(*u).distance+w;
        (*v).pathlength=(*u).pathlength+1;
        (*v).pinedge=pcn;
    }
}

// Queue Operation

static list<sssp_entry>::iterator get_maximum (list<sssp_entry>& sq)
{
    list<sssp_entry>::iterator max, gi=sq.begin();
    max=gi;
    while(gi!=sq.end())
    {
        if((*max).distance<(*gi).distance) max=gi;
        gi++;
    }
    return max;
}

// Find SSSP iterator

static list<sssp_entry>::iterator find_iterator (list<sssp_entry>& sq, TaskType *ptask)
{
    list<sssp_entry>::iterator gi=sq.begin();
    while(gi!=sq.end())
    {
        if((*gi).ptask==ptask) return gi;
        gi++;
    }
    return gi;
}

// Set BFS visited

static void set_bfs_mark (TaskType *t, bool b, list<bfs_tmp_data>& l)
{
    list<bfs_tmp_data>::iterator si;

    for(si=l.begin();si!=l.end();si++)
    {
        if(si->node==t) 
        {
            si->visited=b;
            return;
        }
    }
}

// Get BFS visited

static bool get_bfs_mark (TaskType *t, list<bfs_tmp_data>& l)
{
    list<bfs_tmp_data>::iterator gi;

    for(gi=l.begin();gi!=l.end();gi++)
    {
        if(gi->node==t) return gi->visited;
    }
	return false;
}

// Topo Sort Auxillary function

static void topo_sort_tmp (TaskType *v, list<bfs_tmp_data>&l, 
        list<TaskType *>* sorted)
{
    set_bfs_mark(v,true,l);
    list<ConnectionType *>::iterator tci;
    
    for(tci=v->Get_Out_Edges_Begin();tci!=v->Get_Out_Edges_End();tci++)
    {
        if(!get_bfs_mark((*tci)->Get_To(), l))
        {
            topo_sort_tmp ((*tci)->Get_To(),l,sorted);
        }
    }
    sorted->push_front(v);
}

// Priority Topo Sort

static void priority_topo_sort_tmp (TaskType *v, list<bfs_tmp_data>&l, 
        list<TaskType *>* sorted)
{
    set_bfs_mark(v,true,l);
    list<ConnectionType *>::iterator tci;
    list<TaskType*> children;
    
    for(tci=v->Get_Out_Edges_Begin();tci!=v->Get_Out_Edges_End();tci++)
    {
        children.push_front((*tci)->Get_To());
    }
    while(!children.empty())
    {
        TaskType* ch=get_min_priority(children);
        
        if(!get_bfs_mark(ch, l))
        {
            priority_topo_sort_tmp (ch,l,sorted);
        }
    }
    sorted->push_front(v);
}

static TaskType* get_min_priority (list<TaskType*>& l)
{
    list<TaskType*>::iterator ti,sel=l.end();
    TaskType* pt=NULL;
    
    for(ti=l.begin();ti!=l.end();ti++)
    {
        TaskType* pt1=*ti;
        
        if(!pt) 
        {
            pt=pt1;
            sel=ti;
        }
        else 
        {
            if(pt1->Property()->Get_Priority()<pt->Property()->Get_Priority())
            {
                pt=pt1;
                sel=ti;
            }
        }
    }
    l.erase(sel);
    return pt;
}
