#include "..\stdafx.h"
#include <math.h>
#include <list>
#include <string>
#include "..\Component.h"
#include "..\linux_port\system.h"
#include "..\linux_port\task_allocation.h"
#include "..\linux_port\main_extern.h"
#include "..\linux_port\graphalgo.h"
#include "..\linux_port\first_fit.h"
#include "component_dialog.h"
#include "Analyzer.h"
#include "constants.h"

using namespace std;

// GME Parsing Functions
static int Get_OS_HW_Parameters();
static int Scan_Model();
static int Scan_Meta_Model ();
static double timer_overhead(PlatformConfig, int); // Shige: for timer overhead computation
static float Get_Trans_Time_For_CAN (int, float);
static void Clear_All ();
static bool Is_Input_Task (CBuilderModel*);
static void Set_Global_Property();

/*****************************************************************/

void Component_Task_Map()
{
	component_dialog cd;			//dialog box for GUI
    int r;
	int do_modal_ret; 

    Clear_All();
    r=Get_OS_HW_Parameters();
    if(r==OP_FAIL) return;
    r=Scan_Model();
    if(r==OP_FAIL) return;

	do_modal_ret = cd.DoModal();	// -1 is returned if the dialog box could 
									// not be created
	
	if(do_modal_ret == -1)		 
		AfxMessageBox("Dialog box could not be created!", MB_OK | MB_ICONSTOP);

	return;
}

static void Clear_All ()
{
    sys.Clean();
    grp.Clean();
    meta_grp.Clean();
	main_clear();
    policy.sp=ADVANCED_RMA;
	policy.debug=false;
    allocdone=false;
    mbuff.Empty();
    lout.Empty();
    return;
}

static int Scan_Model ()
{
    int r;

    r=Scan_Meta_Model();
    if(r==OP_FAIL) return OP_FAIL;

    CBuilderModel *root=NULL;
    
	const CBuilderFolder *folder = CBuilder::theInstance->GetFolder(SW_FOLDER);
	if(!folder) {
		AfxMessageBox("No task root folder of name TaskFolder");
		return OP_FAIL;
	}
    
	const CBuilderModelList *rootlist=folder->GetRootModels();
    POSITION modpos = rootlist->GetHeadPosition();
    while(modpos)
    {
        root=rootlist->GetNext(modpos);
        break;
    }
    
	if(!root) {
		AfxMessageBox("No SW Root Model!");
		return OP_FAIL;
	}

    const CBuilderModelList *ltasks=root->GetModels();
    POSITION tpos=ltasks->GetHeadPosition();
    while(tpos)
    {
	    CBuilderModel *model=(CBuilderModel*) ltasks->GetNext(tpos);
	    CString fname=model->GetName();
        TaskType* pmeta_task=meta_grp.Get_Node_By_Name(fname);
		
		// Get CPU ref
	
	    Processor *alloc_proc;
		const CBuilderAtomReferenceList *lref=model->GetAtomReferences("SWCpuRef");

		POSITION lrefpos = lref->GetHeadPosition();
		if(lrefpos)
		{
			CBuilderAtomReference *ref=lref->GetNext(lrefpos);
	        CBuilderAtom *cpu_referred=(CBuilderAtom *) ref->GetReferred();
			CString cpu_name=cpu_referred->GetName();
	        alloc_proc=sys.Get_Proc_By_Name(cpu_name);
	    }
        else
		{
			AfxMessageBox("Enter Processor Reference for "+model->GetName());
			return OP_FAIL;
		}

	    double period;
	    if(!model->GetAttribute("Period",period))
	    {
	        AfxMessageBox("Error getting Period for "+model->GetName());
	        return OP_FAIL;
	    }

	    double deadline;
	    if(!model->GetAttribute("SystemDeadline",deadline))
	    {
	        AfxMessageBox("Error getting Deadline for "+model->GetName());
	        return OP_FAIL;
	    }

	    const CBuilderModelList *models = model->GetModels();
	    
	    POSITION modelPos = models->GetHeadPosition();
		while(modelPos)
		{
			CBuilderModel *tmp_model = models->GetNext(modelPos);
	        CString task_name=fname+"."+tmp_model->GetName();
	
	        // Get Priority
	
			int task_priority;
			if(!tmp_model->GetAttribute("Priority",task_priority))
			{
				AfxMessageBox("Error getting Priority for "+tmp_model->GetName());
				return OP_FAIL;
			}
	
            // Get execution time
            
	        double execution_time;
	        if(!tmp_model->GetAttribute("ExecutionTime",execution_time))
	        {
	            AfxMessageBox("Error getting ExecutionTime for "+tmp_model->GetName());
	            return OP_FAIL;
	        }
	        if(execution_time==0)
	        {
	            AfxMessageBox("Set execution time for "+tmp_model->GetName());
	            return OP_FAIL;
	        }
	
	
            TaskType* ptask=new TaskType;
            TaskProperty* ptp=new TaskProperty;
            TaskAlgoProperty* pap=new TaskAlgoProperty;
	
            ptp->Set_Algo_Property(pap);
            ptp->Set_Task_Pointer(ptask);
            ptask->Set_Property(ptp);
	        ptp->Set_Name(task_name);
            ptp->Set_Task_Name(fname);
            ptp->Set_Exec_Time(execution_time);
            ptp->Set_Alloc_Proc(alloc_proc);
            ptp->Set_User_Priority(task_priority);
			ptp->Set_Priority(task_priority);
            ptp->Set_Meta_Pointer(pmeta_task);
	            
            if(Is_Input_Task(tmp_model))
            {
                TaskGlobalProperty *pgl=new TaskGlobalProperty;
	            pgl->Set_Global_Root(ptask);
	            pgl->Set_Global_Period(period);
	            pgl->Set_Global_Deadline(deadline);
                ptask->Property()->Set_Global_Property(pgl);
				ptask->Property()->Set_Period(period);
                pmeta_task->Property()->Set_Meta_Pointer(ptask);
            }

            grp.Add_Node(ptask);
        }
	
	        /********* Action-Action links within a Task **************/
	
	    modelPos = models->GetHeadPosition();
		while(modelPos)
		{
			CBuilderModel *tmp_model = models->GetNext(modelPos);
	        CString task_name=fname+"."+tmp_model->GetName();
            
            TaskType* pfrom=grp.Get_Node_By_Name(task_name);

	        const CBuilderAtomList *eatoms=tmp_model->GetAtoms("OutEvent");
	        POSITION ePos=eatoms->GetHeadPosition();
	        while(ePos)
	        {
                CBuilderAtom* eatom=(CBuilderAtom*)eatoms->GetNext(ePos);

                double size;
	            if(!eatom->GetAttribute("CommDataSize",size))
    	        {   
	                AfxMessageBox("Error getting CommDataSize for "+eatom->GetName());
	                return OP_FAIL;
	            }
	                    
	            CBuilderObjectList ctl;
	            eatom->GetOutConnectedObjects("SWEventConnection",ctl);
	            if(!ctl.IsEmpty())
	            {
	                POSITION tPos=ctl.GetHeadPosition();
	                while(tPos)
                    {
	                    CBuilderAtom* tatom=(CBuilderAtom*) ctl.GetNext(tPos);
                        CString to_name=tatom->GetParent()->GetName();

                        TaskType* pto=grp.Get_Node_By_Name(fname+"."+to_name);
                        
                        ConnectionType* pc=new ConnectionType;
                        ConnectionProperty* pcp=new ConnectionProperty;
	
                        pcp->Set_Message_Size(size);
                        pcp->Set_Trans_Time(0);
                        pc->Set_Property(pcp);
	                        
                        pc->Set_From(pfrom);
                        pc->Set_To(pto);
                        pfrom->Add_Out_Connection(pc);
                        pto->Add_In_Connection(pc);

                        grp.Add_Edge(pc);
                    }
                }
            }
        }
        
	    Set_Global_Property();
	        
	    /******************** Meta-task priority orderings  *********************/
	    
    }

	return OP_SUCCESS;
}

static int Scan_Meta_Model ()
{
    CBuilderModel *root=NULL;
    
	const CBuilderFolder *folder = CBuilder::theInstance->GetFolder(SW_FOLDER);
	if(!folder) {
		AfxMessageBox("No task root folder of name TaskFolder");
		return OP_FAIL;
	}
    
	const CBuilderModelList *rootlist=folder->GetRootModels();
    POSITION modpos = rootlist->GetHeadPosition();
    while(modpos)
    {
        root=rootlist->GetNext(modpos);
        break;
    }
    
	if(!root) {
		AfxMessageBox("No SW Root Model!");
		return OP_FAIL;
	}

    const CBuilderModelList *ltasks=root->GetModels();
    POSITION tpos=ltasks->GetHeadPosition();
    while(tpos)
    {
	    CBuilderModel *model=(CBuilderModel*) ltasks->GetNext(tpos);
	    CString fname=model->GetName();
		
	    double period;
	    if(!model->GetAttribute("Period",period))
	    {
	        AfxMessageBox("Error getting Period for "+model->GetName());
	        return OP_FAIL;
	    }

	    double deadline;
	    if(!model->GetAttribute("SystemDeadline",deadline))
	    {
	        AfxMessageBox("Error getting Deadline for "+model->GetName());
	        return OP_FAIL;
	    }

        TaskType* ptask=new TaskType;
        TaskProperty* ptp=new TaskProperty;
	
        ptp->Set_Task_Pointer(ptask);
        ptask->Set_Property(ptp);
	    ptp->Set_Name(fname);
        ptp->Set_Period(period);
        ptp->Set_Deadline(deadline);
	            
        meta_grp.Add_Node(ptask);
    }

    ltasks=root->GetModels();
    tpos=ltasks->GetHeadPosition();
    while(tpos)
    {
	    CBuilderModel *model=(CBuilderModel*) ltasks->GetNext(tpos);
	    CString fname=model->GetName();
        TaskType* pfrom=meta_grp.Get_Node_By_Name(fname);
		
        const CBuilderAtomList* datoms=model->GetAtoms("OutData");
        POSITION dPos=datoms->GetHeadPosition();
        while(dPos)
        {
            CBuilderAtom* datom=(CBuilderAtom*) datoms->GetNext(dPos);
	        CBuilderObjectList ctl;
            
	        datom->GetOutConnectedObjects("SWDataConnection",ctl);
	        if(!ctl.IsEmpty())
	        {
	            POSITION tPos=ctl.GetHeadPosition();
	            while(tPos)
                {
	                CBuilderAtom* tatom=(CBuilderAtom*) ctl.GetNext(tPos);
                    CString to_name=tatom->GetParent()->GetName();
                    TaskType* pto=meta_grp.Get_Node_By_Name(to_name);

                    ConnectionType* pc=new ConnectionType;
                    ConnectionProperty* pcp=new ConnectionProperty;
	
                    pcp->Set_Message_Size(0);
                    pcp->Set_Trans_Time(0);
                    pc->Set_Property(pcp);
	                        
                    pc->Set_From(pfrom);
                    pc->Set_To(pto);
                    pfrom->Add_Out_Connection(pc);
                    pto->Add_In_Connection(pc);

                    meta_grp.Add_Edge(pc);
                }
            }
        }
    }

	return OP_SUCCESS;
}

static bool Is_Input_Task (CBuilderModel* pm)
{
    const CBuilderAtomList *eatoms=pm->GetAtoms("InEvent");
	POSITION ePos=eatoms->GetHeadPosition();
	while(ePos)
    {
        CBuilderAtom* eatom=(CBuilderAtom*) eatoms->GetNext(ePos);
	                    
        CBuilderObjectList ctl;
        eatom->GetInConnectedObjects("SWEventConnection",ctl);
        if(!ctl.IsEmpty()) return false;
    }
    return true;
}

static void Set_Global_Property()
{
    list<TaskType*> succlist;
    TaskType* ptask;
    list<TaskType*>::iterator pti,pti2;
    TaskGlobalProperty* pgl;

    for(pti=grp.Node_Begin();pti!=grp.Node_End();pti++)
    {
        ptask=*pti;

        if(ptask->Get_Number_Of_In_Connection()==0)
        {
            pgl=ptask->Property()->Get_Global_Property();
            succlist.clear();
            get_succ_task_list(succlist,ptask);
            
            for(pti2=succlist.begin();pti2!=succlist.end();pti2++)
            {
                TaskType* ptask2=*pti2;

                ptask2->Property()->Set_Global_Property(pgl);
            }
        }
    }
}

static int Get_OS_HW_Parameters()
{
	double overhead; // shige add to record overhead. Only used if no profile available 6/6

    CBuilderModel *rootModel=NULL;
    CBuilderFolder *hwFolder = CBuilder::theInstance->GetFolder(HW_FOLDER);
    
    if (hwFolder==NULL)
    {
        AfxMessageBox("No PlatformFolder for platform configufation model");
        return OP_FAIL;
    }
    
	const CBuilderModelList *rootModels = hwFolder->GetRootModels();
    POSITION modpos=rootModels->GetHeadPosition();
    while(modpos)
    {
        rootModel=rootModels->GetNext(modpos);
        break;
    }

    if(rootModel==NULL)
    {
        AfxMessageBox("No hardware model present!");
        return OP_FAIL;
    }

    // Get OS

    const CBuilderAtomList *osatoms = rootModel->GetAtoms(OS_NAME);
    POSITION ospos=osatoms->GetHeadPosition();
    while(ospos)
    {
        CBuilderAtom *os=(CBuilderAtom *) osatoms->GetNext(ospos);

		/* Comment out. determined based on profiling. Shige 5/29 */
		/* shige revised. Keep this value, and use it when profile is not available. 6/6 */
		if(!os->GetAttribute("TimerOverhead",overhead)) overhead = 0;
        /* end comment. Shige 5/29 */

		/* Add to read timer resolution. Shige 6/1 */
		int res;
		if (!os->GetAttribute("TimerRes", res))
		{
			AfxMessageBox("Error getting Timer Resolution for "+os->GetName()); 
			return OP_FAIL;
		}
		/* end read timer resolution */


		double sh;
		if(!os->GetAttribute("SchedulingOverhead",sh))
		{
			AfxMessageBox("Error getting Scheduling Overhead for "+os->GetName()); 
			return OP_FAIL;
        }
        
		double ch;
		if(!os->GetAttribute("ContextSwitchOverhead",ch))
		{
			AfxMessageBox("Error getting CSW Overhead for "+os->GetName()); 
			return OP_FAIL;
        }

        OS* pos=new OS;
        OperatingSystemProperty* posp=new OperatingSystemProperty;

        posp->Set_Scheduling_Overhead(sh);
        posp->Set_Context_Switch_Overhead(ch);
        posp->Set_Timer_Overhead(0);
        posp->Set_Timer_Resolution(res);
		CString os_name=os->GetName();
        posp->Set_Name(os_name);
        pos->Set_Property(posp);

        sys.Add_OS(pos);
    }
    
    // Get Network

	const CBuilderAtomList *natoms = rootModel->GetAtoms(CANBUS);
	POSITION npos = natoms->GetHeadPosition();
	while(npos)
	{
		CBuilderAtom *net=(CBuilderAtom *) natoms->GetNext(npos);

		Network* pnet=new Network;
		NetworkProperty* pnp=new NetworkProperty;

		pnet->Set_Property(pnp);
		CString net_name=net->GetName();
		pnp->Set_Name(net_name);

	    int mbps;
		if(!net->GetAttribute("NetworkSpeed",mbps))
		{
			AfxMessageBox("Error could not get bit trans time for "+net->GetName());
			return OP_FAIL;
		}

		pnp->Set_Bit_Trans_Time((1.0*1000)/((float)mbps*1000000.0));

		sys.Add_Net(pnet);
	}

    
	const CBuilderAtomList *atoms = rootModel->GetAtoms(CPU);
	POSITION pos = atoms->GetHeadPosition();
	while(pos) 
    {
		CBuilderAtom *ba = (CBuilderAtom *) atoms->GetNext(pos);

        ProcessorProperty* ppp=new ProcessorProperty;
        Processor* pproc=new Processor;

		int canid;
		if(!ba->GetAttribute("CANID",canid))
		{
			AfxMessageBox("Error getting CAN Id for "+ba->GetName());
			return OP_FAIL;
		}

        ppp->Set_ID(canid); // Get CAN ID here

		/* add to get CPU type. Shige 5/29 */
		int hwtype;
		if ( !ba->GetAttribute("CPUType", hwtype) ) 
		{
			AfxMessageBox("Error getting CPU Type for "+ba->GetName()); 
			return OP_FAIL;
        }
		/* end adding. Shige 5/29 */

		double bound;
		if ( !ba->GetAttribute("Bound", bound) ) 
		{
			AfxMessageBox("Error getting Bound for "+ba->GetName()); 
			return OP_FAIL;
        }

		if(bound > 1 || bound < 0)
		{
			AfxMessageBox("Error bound is not in the range of 0 to 1 for "+ba->GetName(), MB_ICONSTOP); 
			return OP_FAIL;
		}

		ppp->Set_Bound(bound);

        ppp->Set_Type(hwtype);
		CString proc_name=ba->GetName();
        ppp->Set_Name(proc_name);
        pproc->Set_Property(ppp);

        // Get OS
        
        CBuilderAtomicObjectList conns;
        ba->GetDirectOutConnections("CpuOSConn", conns);
        if(!conns.IsEmpty())
        {
            POSITION connPos=conns.GetHeadPosition();
            if(connPos)
            {
                CBuilderAtom *dos=(CBuilderAtom *) conns.GetNext(connPos);
				CString dos_name=dos->GetName();
                OS* pos=sys.Get_OS_By_Name(dos_name);
                
    			/* this is added to retrieve platform information. Shige 5/29 */
				int otype;
				if(!dos->GetAttribute("OSType",otype))
				{
					AfxMessageBox("Error getting OS for "+dos_name); 
					return OP_FAIL;
				}
		
				PlatformConfig pconf = UNDECIDE;
				if ( (hwtype == MPC) && (otype == OSEK) ) pconf = MPC_OSEK;
				else if ( (hwtype == INTEL) && (otype == QNX) ) pconf =INTEL_QNX;
				else if ( (hwtype == INTEL) && (otype == RTL) ) pconf = INTEL_RTL;
				else 
                {
                    CString str;
					switch (hwtype) 
                    {
						case MPC: str = "MPC";break;
						case INTEL: str ="INTEL";break;
						default: str = "unknown HW";break;
					}
					CString osstr;
					switch (otype) 
                    {
						case OSEK: osstr = "OSEKWorks";break;
						case QNX: osstr = "QNX";break;
						case RTL: osstr = "RTLinux";break;
						default: osstr = "unknown OS";break;
					}
                    AfxMessageBox("The OS:"+ osstr +" is not profiled on HW:" + str +". Use user-specified value in Timer Overhead instead.");
				    pconf = UNDECIDE;
				}
				pos->Property()->Set_Timer_Overhead(timer_overhead(pconf, pos->Property()->Get_Timer_Resolution()));
                if (pconf == UNDECIDE ) pos->Property()->Set_Timer_Overhead(overhead);
				if (pos->Property()->Get_Timer_Overhead()==-1) pos->Property()->Set_Timer_Overhead(1000/pos->Property()->Get_Timer_Resolution()); // shige: force utilization of timer overhead is 1.
					/* end adding. Shige 5/29 */

                pproc->Set_OS(pos);
            }
        }

        // Get Network

		CBuilderAtomicObjectList nconns;
		ba->GetDirectOutConnections("HWProcNetworkConn",nconns);
		if(!nconns.IsEmpty())
		{
			POSITION connpos=nconns.GetHeadPosition();
			if(connpos)
			{
				CBuilderAtom *nos=(CBuilderAtom*) nconns.GetNext(connpos);
				CString net_name=nos->GetName();

				Network* pnet=sys.Get_Network_By_Name(net_name);
				pnet->Add_Processor(pproc);
				pproc->Add_Net(pnet);
			}
		}

        sys.Add_Proc(pproc);
	}
    return OP_SUCCESS;
}

#define MAX_PROFILED_RES  7

static double timer_overhead(PlatformConfig pconf, int res)
{
	struct { /*define a table to store profiled information*/
		int res;		// unit: ticks/sec
		double value;   // unit: microsecond
	} oh_table[20][MAX_PROFILED_RES];

	double t_oh;

	/* initialize table */
	oh_table[MPC_OSEK][0].res = 10000;
	oh_table[MPC_OSEK][0].value = 13.67/1000; // in ms
	oh_table[MPC_OSEK][1].res = 5000;
	oh_table[MPC_OSEK][1].value = 8.13/1000;
	oh_table[MPC_OSEK][2].res = 2500;
	oh_table[MPC_OSEK][2].value = 6.77/1000;	
	oh_table[MPC_OSEK][3].res = 2000;
	oh_table[MPC_OSEK][3].value = 6.54/1000;
	oh_table[MPC_OSEK][4].res = 1250;
	oh_table[MPC_OSEK][4].value = 6.24/1000;
	oh_table[MPC_OSEK][5].res = 1000;
	oh_table[MPC_OSEK][5].value = 6.15/1000;
	oh_table[MPC_OSEK][6].res = 100;
	oh_table[MPC_OSEK][6].value = 6/1000;

	/* find the overhead of the given resolution */
	for (int i=0; i<MAX_PROFILED_RES; i++) 
		if ( oh_table[pconf][i].res <= res ) break;

	if ( i == 0 ) /* given resolution smaller than min_profiled, system will halt without any application running */
		t_oh = -1;
	if ( i > MAX_PROFILED_RES ) /* given resolution greater than max_profiled, overhead is constant */
		t_oh = oh_table[pconf][MAX_PROFILED_RES].value;
	if ( (i>0) && (i<MAX_PROFILED_RES) ) /* otherwise, need to compute */
		t_oh = oh_table[pconf][i-1].value + (oh_table[pconf][i-1].value - oh_table[pconf][i].value)/(oh_table[pconf][i].res - oh_table[pconf][i-1].res)*res;
	if ( pconf == UNDECIDE ) t_oh = 0;

	return t_oh;  // unit: ms/tick 
}

static float Get_Trans_Time_For_CAN (int s, float bit_time)
{
    return (floor((34.0+8*s)/5)+47+8*s)*bit_time;
}
