#include "..\stdafx.h"
//#include "windows.h"
//#include <stdio.h>
#include <algorithm>
#include "graphalgo.h"
#include "system.h"
#include "treenode.h"
#include "schedule.h"
#include "branch_bound.h"
#include "main_extern.h"
#include "debug.h"

#undef max
#define max(x,y) (((x)>(y))?(x):(y))
#undef min
#define min(x,y) (((x)<(y))?(x):(y))

static float evaluate(TreeNode *,System&,Graph&);
static TreeNode * pop_max_util_node(list<TreeNode *>&);
static bool solution(TreeNode*,System&,Graph&);
static TaskType* get_next_task_to_expand(TreeNode*,System&,Graph&);
static void get_proc_list(list<Processor *>&,TaskType*,TreeNode*,System&,Graph&);
static void cleanup(TreeNode *);
static void cleandown(TreeNode *);
static void clear_procs(System&);
static void clear_tasks(Graph&);
static void get_response_times_tasks(System&,Graph&);
static void get_response_times_messages(System&,Graph&);
static float find_evaluation_function(TreeNode*,System&,Graph&);
static void get_mapping (TreeNode*,System&,Graph&);
static void save_best_and_delete (TreeNode*&, TreeNode*);

bool Branch_Bound_Algorithm (System &s, allocation_stats& as, Graph& g)
{
    TreeNode *root;
    TreeNode *bestsol=NULL;
    list<TreeNode *> queue;
    list<Processor *> cand_procs;
    list<Processor *>::iterator cpi;
    
    root = new TreeNode;
    root->Set_Value(evaluate(root,s,g));
    queue.push_front(root);

    while(!queue.empty())
    {
        TreeNode *node = pop_max_util_node(queue);
        as.Incr();

        if(solution(node, s, g)) 
        {
            save_best_and_delete(bestsol,node);
            continue;
        }
    
        //node is not the solution. so expand one more step
        
        TaskType *next_task=get_next_task_to_expand(node, s, g);

        if(next_task==NULL)
        {
            save_best_and_delete(bestsol,node);
        }
        else
        {
            cand_procs.clear();
            get_proc_list(cand_procs, next_task, node, s, g);
            if(cand_procs.size()==0)
            {
                save_best_and_delete(bestsol,node);
            }
            else
            {
                for(cpi=cand_procs.begin();cpi!=cand_procs.end();cpi++)
                {
                    Processor *pproc=*cpi;
    
                    TreeNode *new_node = new TreeNode;
                    new_node->Set_Task(next_task);
                    new_node->Set_Processor(pproc);
                    new_node->Set_Parent(node);
                    node->Add_Child(new_node);
                    new_node->Clear_Children();
                    new_node->Set_Value(evaluate(new_node,s,g));
                    queue.push_front(new_node);
                }
            }
        }
    }

    if(bestsol)
    {
        bool ret=bestsol->Get_Solution();
        evaluate(bestsol,s,g);
        cleandown(root);
        return ret;
    }
    else
    {
        cleandown(root);
        return false;
    }
}

float Evaluate(System& s, Graph& g)
{
    get_response_times_tasks(s, g);
    get_response_times_messages(s, g);

    return find_evaluation_function (NULL, s, g);
}

void save_response_times(list<old_exe_times>& srt, Graph& g)
{
    list<TaskType*>::iterator si;
    list<ConnectionType*>::iterator ci;

    for(si=g.Node_Begin();si!=g.Node_End();si++)
    {
        TaskType *ptt=*si;
        old_exe_times ort;
    
        ort.pt=ptt;
        ort.istask=true;
        ort.exe_time=ptt->Property()->Get_Exec_Time();
        srt.push_front(ort);
    }
        
    for(ci=g.Edge_Begin();ci!=g.Edge_End();ci++)
    {
        ConnectionType *pc=*ci;
        old_exe_times ort;

        ort.pc=pc;
        ort.istask=false;
        ort.exe_time=pc->Property()->Get_Trans_Time();
        srt.push_front(ort);
    }

    return;
}

void restore_response_times(list<old_exe_times>& srt)
{
    TaskType* pt;
    ConnectionType* pc;
    list<old_exe_times>::iterator ri;

    for(ri=srt.begin();ri!=srt.end();ri++)
    {
        old_exe_times ort=*ri;

        if(ort.istask)
        {
            pt=ort.pt;
            pt->Property()->Set_Exec_Time(ort.exe_time);
        }
        else
        {
            pc=ort.pc;
            pc->Property()->Set_Trans_Time(ort.exe_time);
        }
    }
}

static TreeNode * pop_max_util_node(list<TreeNode *>& queue)
{
    list<TreeNode *>::iterator pi,si;
    TreeNode* ptn=NULL;
    float max_value=-MAX;
    
    for(pi=queue.begin();pi!=queue.end();pi++)
    {
        TreeNode* pt=*pi;

        if(pt->Get_Value()>max_value) 
        {
            ptn=pt;
            si=pi;
        }
    }
    if(ptn) queue.erase(si);
    return ptn;
}

static void get_mapping (TreeNode *tn, System&s, Graph& g)
{
    TreeNode *ptmp;
    
    clear_procs(s);
    clear_tasks(g);

    ptmp=tn;
    while(ptmp!=NULL)
    {
        TaskType *ptask=ptmp->Get_Task();
        Processor *pproc=ptmp->Get_Processor();

        if((ptask)&&(pproc))
        {
            pproc->Add_Task(ptask);
            ptask->Property()->Set_Proc(pproc);
        }
        ptmp=ptmp->Get_Parent();
    }
    return;
}

static float evaluate(TreeNode * tn, System& s, Graph& g)
{
    
    get_mapping(tn,s,g);


    get_response_times_tasks(s, g);
    get_response_times_messages(s, g);

    return find_evaluation_function (tn, s, g);
}

static void get_response_times_tasks (System& s, Graph& g)
{
    list<Processor *>::iterator ip;

    for(ip=s.Proc_Begin();ip!=s.Proc_End();ip++)
    {
        Processor *pproc=*ip;

        Schedule_Task(pproc);
    }
    return;
}

static void get_response_times_messages (System& s, Graph& g)
{
    /* Get messages on the network */

    list<Processor *>::iterator ip;
    for(ip=s.Proc_Begin();ip!=s.Proc_End();ip++)
    {
        Processor *pproc=*ip;
        list<TaskType *>::iterator it;

        for(it=pproc->Tasks_Begin();it!=pproc->Tasks_End();it++)
        {
            TaskType* ptask=*it;

            list<ConnectionType *>::iterator ic;

            for(ic=ptask->Get_Out_Edges_Begin();ic!=ptask->Get_Out_Edges_End();ic++)
            {
                ConnectionType *pc=*ic;
                
                Processor *other_proc=pc->Get_To()->Property()->Get_Proc();
                if((other_proc)&&(other_proc!=pproc))
                    pproc->Add_Connection(pc);
            }
        }
    }

    list<ConnectionType*>::iterator ic;
    
    for(ic=g.Edge_Begin();ic!=g.Edge_End();ic++)
    {
        ConnectionType* pc=*ic;

        pc->Property()->Set_Response_Time(0);
    }

    Schedule_CAN_Messages(s);

    return;
}

static float find_evaluation_function (TreeNode* tn, System& s, Graph& g)
{
    list<Processor *>::iterator ip;
    list<old_exe_times> ort;
    list<TaskType*>::iterator it;
    list<ConnectionType*>::iterator ic;
    float min_task_delay=MAX;
    float min_message_delay=MAX;
    float min_end_to_end_delay=MAX;
    bool tasks_remaining=false;
    int num_tasks_scheduled, num_message_scheduled;

    num_tasks_scheduled=num_message_scheduled=0;

    save_response_times(ort, g);

    for(it=g.Node_Begin();it!=g.Node_End();it++)
    {
        TaskType* ptask=*it;
        float resp_time;

        if(ptask->Property()->Get_Proc())
        {
            resp_time=ptask->Property()->Get_Response_Time();
            ptask->Property()->Set_Exec_Time(resp_time);
            if(resp_time<ptask->Property()->Get_Period())
                num_tasks_scheduled++;
        }
        else
        {
            // Here is where we can tighten the strings on
            // bounds.

            tasks_remaining=true;
            resp_time=ptask->Property()->Get_Exec_Time(); // Most optimistic. 
        }
        
        float delay=ptask->Property()->Get_Period()-resp_time;
        if(delay<min_task_delay) min_task_delay=delay;
    }

    for(ic=g.Edge_Begin();ic!=g.Edge_End();ic++)
    {
        ConnectionType* pc=*ic;

        float resp_time;

        if(pc->Get_From()->Property()->Get_Proc()&&
                pc->Get_To()->Property()->Get_Proc())
        {
            resp_time=pc->Property()->Get_Response_Time();
            pc->Property()->Set_Trans_Time(resp_time);
            if(resp_time<pc->Get_To()->Property()->Get_Period())
                num_message_scheduled++;
        }
        else
        {
            // A procedure to estimate maximum resp_time for
            // this message needed.
            
            resp_time=0;
            pc->Property()->Set_Trans_Time(resp_time);
        }
    
        float delay=pc->Get_From()->Property()->Get_Period()-resp_time;
        if(delay<min_message_delay) min_message_delay=delay;
    }
    
    for(it=g.Node_Begin();it!=g.Node_End();it++)
    {
        TaskType* ptask=*it;
        list<sssp_entry>* pss;
        list<sssp_entry>::iterator isp;
        float max_delay=0;
        
        if(ptask->Get_Number_Of_In_Connection()!=0) continue;

        pss=SSSP_Bellman_Ford(g, *ptask);
        for(isp=pss->begin();isp!=pss->end();isp++)
        {
            sssp_entry se=*isp;

			if(se.ptask->Get_Number_Of_Out_Connection()==0)
			{
				float delay=se.ptask->Property()->Get_Global_Deadline()-se.distance;
				if(delay<min_end_to_end_delay) min_end_to_end_delay=delay;
			}

        }
        delete pss;
    }

    restore_response_times(ort);

	float ret_val=min(min(min_task_delay,min_message_delay), min_end_to_end_delay);
	if((!tasks_remaining)&&(ret_val)&&(tn))
        tn->Set_Solution(true);
	return ret_val;
}
    
static void clear_procs(System& s)
{
    list<Processor *>::iterator pi;

    for(pi=s.Proc_Begin();pi!=s.Proc_End();pi++)
    {
        Processor *pproc=*pi;

        pproc->Clear_Tasks();
        pproc->Clear_Connections();
    }
    return;
}

static void clear_tasks(Graph& g)
{
    list<TaskType *>::iterator ci;

    for(ci=g.Node_Begin();ci!=g.Node_End();ci++)
    {
        TaskType *pt=*ci;

        pt->Property()->Set_Proc(NULL);
    }
}

/*
* Very very stupid procedure returning next task to be
* scheduled. A good hueristic will also consider timing
* constraints or message communication of latency while
* doing this allocation.
*/

static TaskType* get_next_task_to_expand(TreeNode* tn, System& s, Graph& g)
{
    list<TaskType*>::iterator it;
    list<TaskType*> unallocated;
    
    get_mapping(tn,s,g);
    
    for(it=g.Node_Begin();it!=g.Node_End();it++)
    {
        TaskType* pt=*it;

        if(pt->Property()->Get_Proc()==NULL)
            unallocated.push_front(pt);
    }

    float max_util=-1;
    TaskType* ret_task=NULL;
    
    for(it=unallocated.begin();it!=unallocated.end();it++)
    {
        TaskType* pt=*it;

        float util=pt->Property()->Get_Exec_Time()/pt->Property()->Get_Period();
        if(util>max_util)
        {
            max_util=util;
            ret_task=pt;
        }
    }
    return ret_task;
}

/*
* Very very stupid procedure returning all processors in the
* system. Can be optimized a whole lot.
*/

static void get_proc_list(list<Processor *>& proc_list, TaskType* pt ,TreeNode* tn, 
        System& s, Graph& g)
{
    list<Processor *>::iterator ip;
	Processor* alloc_proc=pt->Property()->Get_Alloc_Proc();

	if(alloc_proc) 
	{
		// already allocated 

		proc_list.push_back(alloc_proc);
		return;
	}

    for(ip=s.Proc_Begin();ip!=s.Proc_End();ip++)
    {
        Processor *pp=*ip;

        proc_list.push_front(pp);
    }
    return;
}

static void cleanup(TreeNode * tn)
{
    TreeNode *parent;

    if(tn==NULL) return;
    parent=tn->Get_Parent();
    if(tn->Get_Number_Of_Children()==0)
    {
        if(parent) parent->Delete_Child(tn);
        delete tn;
        cleanup(parent);
    }
    return;
}

static bool solution(TreeNode* tn,System& s,Graph& g)
{
    return tn->Get_Solution();
}

static void cleandown(TreeNode *tn)
{
    list<TreeNode*> children;
    list<TreeNode*>::iterator ic;
    
    if(tn==NULL) return;

    for(ic=tn->Begin_Children();ic!=tn->End_Children();ic++)
    {
        children.push_front(*ic);
    }

    delete tn;

    for(ic=children.begin();ic!=children.end();ic++)
        cleandown(*ic);
 
    return;
}

static void save_best_and_delete (TreeNode*& bestsol, TreeNode* node)
{
    TreeNode *delnode=node;

    if(node==NULL) return;
    
    if(bestsol)
    {
        if(bestsol->Get_Solution())
        {
            if(node->Get_Solution()&&(node->Get_Value()>bestsol->Get_Value()))
            {
                delnode=bestsol;
                bestsol=node;
            }
        }
        else
        {
            if(node->Get_Solution())
            {
                delnode=bestsol;
                bestsol=node;
            }
            else
                if(node->Get_Value()>bestsol->Get_Value())
                {
                    delnode=bestsol;
                    bestsol=node;
                }
        }
    }
    else
    {
        bestsol=node;
        delnode=NULL;
    }
    cleanup(delnode);
}
    
