#include "load_balance.h"

//global includes
#include <set>

//local includes
#include "graph.h"
#include "system.h"
#include "debug.h"
#include "first_fit.h"	//to obtain the struct task_greater_than

using namespace std;

//externals
extern Graph grp;
extern System sys;

double Do_LoadBalance_Mapping()
{
	multiset<TaskType*, task_greater_than>		tasks;
	multiset<Processor*, processor_less_than>	processors;
	double										non_allocated_utility = 0;

	//reset all processors
	for(list<Processor *>::iterator proc_it = sys.Proc_Begin(); proc_it != sys.Proc_End(); ++proc_it)
	{
		Processor* proc_ptr = *proc_it;

		proc_ptr->Clear_Tasks();							//clear all tasks on the processor
		proc_ptr->Property()->Reset_Available_Space();		//give the processor its entire amount of space available
		processors.insert(proc_ptr);						//add into processors set
	}

	//automatically sorts in order of the utilization
	for(list<TaskType*>::iterator grp_it = grp.Node_Begin(); grp_it != grp.Node_End(); grp_it++)
		tasks.insert(*grp_it);

	//map each task to a processor
	for(multiset<TaskType*, task_greater_than>::iterator task_it = tasks.begin(); task_it != tasks.end(); ++task_it)
	{
		double utilization = (*task_it)->Property()->Get_Exec_Time() / (*task_it)->Property()->Get_Period();;

		//determine greatest difference in total utilization for processors
		for(multiset<Processor*, processor_less_than>::iterator proc_it = processors.begin(); proc_it != processors.end(); ++proc_it)
		{
			//is there room for this task on this processor
			if((*proc_it)->Property()->Allocate_Space(utilization))
			{
				Processor* proc_ptr = *proc_it;

				//need to remove old value from the multiset then reinsert it to keep the 
				//multiset sorted
				processors.erase(proc_it);
				proc_ptr->Add_Task(*task_it);
				processors.insert(proc_ptr);
				break;
			}
		}

		//could not find a processor to map the task to
		if(proc_it == processors.end())
			non_allocated_utility += utilization;
	}

	//were there any tasks that could not be mapped to a processor (error)
	if(non_allocated_utility)
		return non_allocated_utility;
	
	//print out the results
	print_LoadBalance_Mapping();

	return 0;	//no errors
}

//Function object definitions
bool processor_less_than::operator () (Processor* p1, Processor* p2)
{
	double utilization1 = p1->Property()->Get_Bound() - p1->Property()->Get_Available_Space();
	double utilization2 = p2->Property()->Get_Bound() - p2->Property()->Get_Available_Space();

	return utilization1 < utilization2;
}
