#ifndef PROCESSOR_NETWORK_H_0
#define PROCESSOR_NETWORK_H_0

#include "graph.h"
#include "property.h"

class OS
{
    private:
        OperatingSystemProperty *property;
    public:
        OS(){property=NULL;};
        ~OS(){if(property) delete property;}
        inline void Set_Property(OperatingSystemProperty *p){property=p;}
        inline OperatingSystemProperty* Get_Property(){return property;}
        inline OperatingSystemProperty* Property(){return property;}
};

class Network
{
    private:
        list<Processor *>  processor_list;
        NetworkProperty *property;
    public:
        Network(){property=NULL;}
        ~Network(){if(property) delete property;}
        inline void Add_Processor (Processor *p) {processor_list.push_front(p);}
        inline void Set_Property(NetworkProperty *p){property=p;}
        inline NetworkProperty* Get_Property(){return property;}
        inline NetworkProperty* Property(){return property;}
};

class Processor
{
    private:
        list<TaskType *> task_list;
        list<ConnectionType *> message_list;
        Network *network;
        ProcessorProperty* property;
        OS *os;
    public:
        Processor(){property=NULL;os=NULL;};
        ~Processor(){if(property) delete property;}
        inline void Add_Task (TaskType *t) {task_list.push_front(t);};
        inline void Add_Network (Network *n) {network=n;};
        inline void Add_Net (Network *n) {network=n;};
        inline void Set_Property (ProcessorProperty *p) {property=p;};
        inline ProcessorProperty* Property(){return property;};
        inline ProcessorProperty* Get_Property(){return property;};
        inline void Set_OS(OS *o){os=o;}
        inline OS* Get_OS(){return os;}
        inline void Clear_Tasks() {task_list.clear();}
        inline void Clear_Connections() {message_list.clear();}
        inline list<TaskType*>::iterator Tasks_Begin() {return task_list.begin();}
        inline list<TaskType*>::iterator Tasks_End() {return task_list.end();}
        inline int Get_Number_Of_Tasks() {return task_list.size();}
        inline void Add_Connection (ConnectionType *c) {message_list.push_front(c);}
        inline list<ConnectionType*>::iterator Connection_Begin()
        {
            return message_list.begin();
        }
        inline list<ConnectionType*>::iterator Connection_End()
        {
            return message_list.end();
        }

};

#endif
