#ifndef PROPERTY_H_0
#define PROPERTY_H_0

#include <list>
#include <set>
#include <afx.h>

using namespace std;

#define NIL         -1
#define MAX         ((float)429496729.0)
#define INVALID(x)  ((x==NIL)||(x==MAX))

class TaskRTProperty
{
    protected:
        float exe_time;
        float period;
        float offset;
        float deadline;
        float response_time;
        float saved_time;
        int priority;
        int user_priority;
        int csw;
        bool period_given;
    public:
        TaskRTProperty()
        {
            exe_time=period=offset=deadline=(float)NIL;
            csw=0;
            response_time=MAX;
            period_given=false;
            priority=user_priority=NIL;
		};
        inline float Get_Exec_Time(){return exe_time;};
        inline float Get_Period(){return period;};
        inline float Get_Offset() {return offset;};
        inline float Get_Deadline(){return deadline;};
        inline float Get_Response_Time(){return response_time;};
        inline int Get_Priority(){return priority;};
        inline void Set_Exec_Time(float f){exe_time=f;return;};
        inline void Set_Period(float f){period=f;return;};
        inline void Set_Offset(float f){offset=f;return;};
        inline void Set_Deadline(float f){deadline=f;return;};
        inline void Set_Response_Time (float f) {response_time=f;return;};
        inline void Set_Priority (int f) {priority=f;return;};
        inline void Reset_Context_Switches(){csw=0;}
        inline void Set_Context_Switches(int i){csw=i;}
        inline int Get_Context_Switches(){return csw;}
        inline void Set_Period_Given(){period_given=true;}
        inline bool Get_Period_Given(){return period_given;}
        inline void Save_Response_Time(){saved_time=response_time;}
        inline void Restore_Response_Time(){response_time=saved_time;}
        inline float Get_User_Priority(){return user_priority;}
        inline void Set_User_Priority(int p){user_priority=p;}
        void Debug();
};

class TaskType;

class TaskIDProperty
{
    private:
        int id;
        TaskType* ptask;
        TaskType* pmeta_task;
		CString name;
        CString task_name;
    public:
        TaskIDProperty(){id=NIL;name="";ptask=NULL;pmeta_task=NULL;};
        inline int Get_ID () {return id;}
        inline void Set_ID (int i) {id=i;return;}
        inline void Set_Name (CString& n) {name=n;}
        inline const CString& Get_Name() {return name;}
        inline const CString& Get_Task_Name () {return task_name;}
        inline void Set_Task_Name (CString& n) {task_name=n;}
        inline void Set_Task_Pointer(TaskType* p){ptask=p;}
        inline void Set_Meta_Pointer(TaskType* p){pmeta_task=p;}
        inline TaskType* Get_Meta_Pointer () {return pmeta_task;}
        inline TaskType* Task(){return ptask;}
};

class ConnectionRTProperty
{
    protected:
        bool type;
        int size;
        int priority;
        float trans_time;
        float response_time;
    public:
        ConnectionRTProperty(){size=0;priority=0;trans_time=0;type=true;};
        inline int Get_Message_Size(){return size;};
        inline int Get_Message_priority(){return priority;};
        inline float Get_Trans_Time(){return trans_time;};
        inline void Set_Message_Size(int s){size=s;return;};
        inline void Set_Message_Priority(int p){priority=p;return;};
        inline void Set_Trans_Time(float t){trans_time=t;return;};
        inline void Set_Response_Time(float t){response_time=t;};
        inline float Get_Response_Time(){return response_time;};
        inline bool Is_Control_Flow(){return type;};
        inline void Set_Data_Flow(){type=false;};
        void Debug();
};

class TaskGlobalProperty
{
    private:
        TaskType* root;
        float exe_time;
        set<TaskType*> SP_private;
        set<TaskType*> LP_private;
        float block_time_private;
        float deadline;
        float period;
    public:
        set<TaskType*> SP;
        set<TaskType*> LP;
        float block_time;
        list<TaskType*> Topo_Order;
    public:
        TaskGlobalProperty()
        {
            SP.clear();exe_time=0;LP.clear();LP_private.clear();
            SP_private.clear();block_time_private=0;block_time=0;
        }
        inline void Set_Global_Root(TaskType *p) {root=p;}
        inline TaskType* Get_Global_Root(){return root;}
        inline float Get_Global_Deadline(){return deadline;}
        inline void Set_Global_Deadline(float f){deadline=f;}
        inline float Get_Global_Period(){return period;};
        inline void Set_Global_Period(float f){period=f;return;};
        int Get_Min_Priority ();
        int Get_Max_Priority ();
        float Get_Global_Response_Time ();
        void Set_Topo_Order ();
        void Canonize();
        void Set_Global_Exe_Time();
        inline float Get_Global_Exe_Time(){return exe_time;}
        inline void Save_SP () 
        {
            SP_private.clear();SP_private=SP;block_time_private=block_time;
            LP_private.clear();LP_private=LP;
        }
        inline bool Changed_SP () 
        {
            if((LP_private.size()==0)||(SP_private.size()==0))
                return true;
            else return 
                ((SP_private!=SP)||(block_time!=block_time_private))
                    ||(LP_private!=LP);
        }
        inline void Restore ()
        {
            SP.clear();SP=SP_private;block_time=block_time_private;
            LP.clear();LP=LP_private;
            return;
        }
        
};

class ProcessorProperty
{
    private:
        int id;
        CString name;
        int type;
		double bound;
		double available_space;
    public:
        ProcessorProperty(){id=NIL;type=NIL;name="";};
        inline void Set_ID(int i){id=i;}
        inline int Get_ID(){return id;}
        inline int Get_CAN_ID(){return id;}
        inline void Set_Type(int t){type=t;}
        inline int Get_Type(){return type;}
        inline void Set_Name(CString& n){name=n;}
        inline const CString& Get_Name(){return name;}
		inline double Get_Bound() {return bound;}
		inline double Get_Available_Space() {return available_space;}
		inline void Reset_Available_Space() {available_space = bound;}
		void Set_Bound(double b);
		bool Allocate_Space(double utility);	//try to allocate space for a task
												//returns true if function call was
												//successful (utility < available space)
};

class NetworkProperty
{
    private:
        float bit_transmission_time;
        CString name;
    public:
        NetworkProperty(){bit_transmission_time=0;name="";};
        inline void Set_Bit_Trans_Time(float f){bit_transmission_time=f;};
        inline float Get_Bit_Trans_Time(){return bit_transmission_time;};
        inline void Set_Name(CString& n){name=n;}
        inline CString& Get_Name(){return name;}
};

class OperatingSystemProperty
{
    private:
        float sched_overhead;
        float context_switch_overhead;
        float timer_overhead;
        float timer_resolution;
        CString name;
    public:
        OperatingSystemProperty()
        {
            sched_overhead=context_switch_overhead=timer_overhead=timer_resolution=0;
            name="";
        }
        inline void Set_Scheduling_Overhead (float f){sched_overhead=f;}
        inline void Set_Context_Switch_Overhead (float f){context_switch_overhead=f;}
        inline void Set_Timer_Overhead(float f){timer_overhead=f;}
        inline void Set_Timer_Resolution(float f){timer_resolution=f;}
        inline void Set_Name(CString& n){name=n;}
        inline float Get_Scheduling_Overhead(){return sched_overhead;}
        inline float Get_Context_Switch_Overhead(){return context_switch_overhead;}
        inline float Get_Timer_Overhead(){return timer_overhead;}
        inline float Get_Timer_Resolution(){return timer_resolution;}
        inline const CString& Get_Name(){return name;}
};

class Processor;

class TaskAlgoProperty
{
    private:
        Processor *proc, *alloc_proc;
        set<TaskType*> MP_private;
    public:
        set<TaskType*> MP;
    public:
        TaskAlgoProperty(){proc=NULL;alloc_proc=NULL;MP_private.clear();MP.clear();}
        void Initialize(){};
        inline void Set_Proc(Processor *p) {proc=p;}
        inline Processor* Get_Proc(){return proc;}
		inline void Set_Alloc_Proc(Processor *p){alloc_proc=p;}
		inline Processor* Get_Alloc_Proc(){return alloc_proc;}
        inline void Save_MP(){MP_private.clear();MP_private=MP;}
        inline bool Changed_MP()
        {
            if(MP_private.size()==0) return true;
            else return MP_private!=MP;
        }
        inline void Restore () {MP.clear();MP=MP_private;}
};

class TaskProperty:public TaskRTProperty, public TaskIDProperty
{
    private:
        TaskGlobalProperty *pglobal;
        TaskAlgoProperty *palgo;
    public:
        TaskProperty(){pglobal=NULL;palgo=NULL;};
        ~TaskProperty(){if(palgo) delete palgo;};
        inline void Set_Global_Property (TaskGlobalProperty *p) {pglobal=p;}
        inline TaskGlobalProperty* Get_Global_Property () {return pglobal;}
        inline void Set_Algo_Property(TaskAlgoProperty *p) {palgo=p;}
        inline float Get_Global_Deadline()
        {
            return (pglobal)? pglobal->Get_Global_Deadline():NIL;
        }
        inline void Set_Proc(Processor *p)
        {
            if(palgo)palgo->Set_Proc(p);
        }
        inline Processor * Get_Proc()
        {
            return palgo? (palgo->Get_Proc()):0;
        }
        inline void Set_Global_Root(TaskType *p) 
        {
            if(pglobal) pglobal->Set_Global_Root(p);
        }
        inline TaskType* Get_Global_Root()
        {
            return (pglobal)? pglobal->Get_Global_Root():NULL;
        }
		inline void Set_Alloc_Proc(Processor *p)
		{
			if(palgo) palgo->Set_Alloc_Proc(p);
		}
		inline Processor* Get_Alloc_Proc ()
		{
			return (palgo)?palgo->Get_Alloc_Proc():NULL;
		}
		/********** Specialization of Period ***********************/
		inline float Get_Period()
		{
			return (pglobal)? pglobal->Get_Global_Period():TaskRTProperty::Get_Period();
		}
		/********** Specialization of Deadline ***********************/
        inline float Get_Deadline()
        {
			return (pglobal)? pglobal->Get_Global_Deadline():0;
		}

        inline float Get_Global_Period ()
        {
            return (pglobal)? pglobal->Get_Global_Period():0;
        }
        int Get_Context_Switches();
        inline set<TaskType*>* MP()
        {
            if(palgo) return &(palgo->MP);
            else return NULL;
        }
        inline set<TaskType*>* SP()
        {
            if(pglobal) return &(pglobal->SP);
            else return NULL;
        }
        inline set<TaskType*>* LP()
        {
            if(pglobal) return &(pglobal->LP);
            else return NULL;
        }
        inline float Get_Global_Response_Time ()
        {
            if(pglobal) return pglobal->Get_Global_Response_Time();
            else return NIL;
        }
        inline int Get_Min_Priority ()
        {
            if(pglobal) return pglobal->Get_Min_Priority();
            else return NIL;
        }
        inline int Get_Max_Priority ()
        {
            if(pglobal) return pglobal->Get_Max_Priority();
            else return NIL;
        }
        int Get_First_Priority();
        int Get_Last_Priority();
        inline list<TaskType*>* Topo()
        {
            if(pglobal) return &(pglobal->Topo_Order);
            else return NULL;
        }
        inline void Set_Topo_Order ()
        {
            if(pglobal) pglobal->Set_Topo_Order();
        }
		inline void Canonize()
		{
			if(pglobal) pglobal->Canonize();
		}
        inline float Get_Global_Exe_Time ()
        {
            if(pglobal) return pglobal->Get_Global_Exe_Time();
            else return 0;
        }
        inline void Set_Global_Exe_Time()
        {
            if(pglobal) pglobal->Set_Global_Exe_Time();
        }
        inline void Save_MP ()
        {
            if(palgo) palgo->Save_MP();
            TaskRTProperty::Save_Response_Time();
        }
        inline void Save_SP ()
        {
            if(pglobal) pglobal->Save_SP();
        }
        inline bool Changed_SP ()
        {
            if(pglobal) return pglobal->Changed_SP();
            else return false;
        }
        inline bool Changed_MP ()
        {
            if(palgo) return palgo->Changed_MP();
            else return false;
        }
        inline void Restore_MP()
        {
            if(palgo) palgo->Restore();
            TaskRTProperty::Restore_Response_Time();
        }
        inline void Restore_SP()
        {
            if(pglobal) pglobal->Restore();
        }
        inline void Set_Block_Time (float& f)
        {
            if(pglobal) { pglobal->block_time=f; }
        }
        inline float Get_Block_Time ()
        {
            if(pglobal) return pglobal->block_time;
            else return (float)99999999;
        }
        bool Changed();
        void Restore();
        void Restore_User_Priorities();
        TaskType* Get_Predecessor();
};

class ConnectionProperty:public ConnectionRTProperty
{
    private:

    public:
        ConnectionProperty(){};
};

#endif

